# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Module defining the base class for general grids (GGD)."""

from __future__ import annotations

from abc import abstractmethod
from typing import Literal

import matplotlib.axes
import numpy as np
from numpy.typing import ArrayLike, NDArray
from raysect.core.math.function.float import Function2D, Function3D
from raysect.core.math.function.vector3d.function2d import Function2D as VectorFunction2D
from raysect.core.math.function.vector3d.function3d import Function3D as VectorFunction3D
from raysect.core.math.vector import Vector3D

ZEROVECTOR = Vector3D(0, 0, 0)


__all__ = ["GGDGrid"]


class GGDGrid:
    """Base class for general grids (GGD).

    Parameters
    ----------
    name
        Name of the grid.
    dimension
        Grid dimensions, by default 1.
    coordinate_system
        Coordinate system, by default ``"cartesian"``.
    """

    def __init__(
        self,
        name: str = "",
        dimension: int = 1,
        coordinate_system: Literal["cylindrical", "cartesian"] = "cartesian",
    ):
        dimension = int(dimension)
        if dimension < 1:
            raise ValueError("Attribute dimension must be >= 1.")

        self._dimension: int = dimension
        self._name: str = str(name)
        self._coordinate_system: str = str(coordinate_system)

        self._interpolator: object
        self._cell_centre: NDArray[np.float64]
        self._cell_area: NDArray[np.float64]
        self._cell_volume: NDArray[np.float64]
        self._mesh_extent: dict[str, float]
        self._num_cell: int

        self._initial_setup()

    @abstractmethod
    def _initial_setup(self) -> None:
        raise NotImplementedError("To be defined in subclass.")

    @property
    def name(self) -> str:
        """Grid name."""
        return self._name

    @name.setter
    def name(self, value: str):
        self._name = str(value)

    @property
    def dimension(self) -> int:
        """Grid dimension."""
        return self._dimension

    @property
    def num_cell(self) -> int:
        """Number of grid cells."""
        return self._num_cell

    @property
    def coordinate_system(self) -> str:
        """Coordinate system."""
        return self._coordinate_system

    @property
    def cell_centre(self) -> NDArray[np.float64]:
        """Coordinate of cell centres as ``(num_cell, dimension)`` array."""
        return self._cell_centre

    @property
    def cell_area(self) -> NDArray[np.float64]:
        """Cell areas as ``(num_cell,)`` array."""
        return self._cell_area

    @property
    def cell_volume(self) -> NDArray[np.float64]:
        """Cell volume as ``(num_cell,)`` array."""
        return self._cell_volume

    @property
    def mesh_extent(self) -> dict[str, float]:
        """Extent of the mesh.

        A dictionary with xmin, xmax, ymin and ymax, ... keys.
        """
        return self._mesh_extent

    @abstractmethod
    def subset(self, indices: ArrayLike, name: str | None = None) -> GGDGrid:
        """Create a subset grid from this instance.

        Parameters
        ----------
        indices
            Indices of the cells of the original grid in the subset.
        name
            Name of the grid subset. Default is ``instance.name + " subset"``.

        Returns
        -------
        GGDGrid
            Subset grid instance.
        """
        raise NotImplementedError("To be defined in subclass.")

    @abstractmethod
    def interpolator(
        self, grid_data: NDArray[np.float64], fill_value: float = 0.0
    ) -> Function2D | Function3D:
        """Return an Function interpolator instance for the data defined on this grid.

        On the second and subsequent calls, the interpolator is created as an instance of the
        previously created interpolator.

        Parameters
        ----------
        grid_data
            Array containing data in the grid cells.
        fill_value
            A value returned outside the grid, by default is 0.0.

        Returns
        -------
        `Function2D` | `Function3D`
            Interpolator instance.
        """
        raise NotImplementedError("To be defined in subclass.")

    @abstractmethod
    def vector_interpolator(
        self, grid_vectors: NDArray[np.float64], fill_vector: Vector3D = ZEROVECTOR
    ) -> VectorFunction2D | VectorFunction3D:
        """Return a VectorFunction interpolator instance for the vector data defined on this grid.

        On the second and subsequent calls, the interpolator is created as an instance of the
        previously created interpolator.

        Parameters
        ----------
        grid_vectors
            ``(3, num_cell)`` Array containing 3D vectors in the grid cells.
        fill_vector
            3D vector returned outside the grid, by default ``Vector3D(0, 0, 0)``.

        Returns
        -------
        `VectorFunction2D` | `VectorFunction3D`
            Interpolator instance.
        """
        raise NotImplementedError("To be defined in subclass.")

    def plot_mesh(
        self, data: NDArray[np.float64] | None = None, ax: matplotlib.axes.Axes | None = None
    ):
        """Plot the grid geometry to a matplotlib figure.

        Parameters
        ----------
        data
            Data array defined on the grid.
        ax
            Matplotlib axes to plot on. If None, a new figure and axes are created.
        """
        raise NotImplementedError("To be defined in subclass.")



"""Module defining simple interpolators for data defined on a 2D unstructured grid."""

from numpy import float64, int32
from numpy.typing import ArrayLike, NDArray
from raysect.core.math import Vector3D
from raysect.core.math.function.float import Function2D
from raysect.core.math.function.float.function2d.interpolate.common import MeshKDTree2D
from raysect.core.math.function.vector3d import Function3D as VectorFunction3D

ZERO_VECTOR = Vector3D(0, 0, 0)

class UnstructGridFunction2D(Function2D):
    """Simple interpolator for the data defined on the 2D unstructured grid.

    Find the cell containing the point (x, y) using the KDtree algorithm.
    Return the data value for this cell or the `fill_value` if the grid does not contain the point.

    Parameters
    ----------
    vertex_coords : (N,3) array_like
        2D array-like with the vertex coordinates of triangles.
    triangles : (M,3) array_like
        2D integer array-like with the vertex indices forming the triangles.
    triangle_to_cell_map : (M,) array_like
        1D integer array-like with the indices of the grid cells (polygons) containing the
        triangles.
    grid_data : (L,) ndarray
        Array containing data in the grid cells.
    fill_value : float, optional
        Value returned outside the gird, by default 0.
    """

    _kdtree: MeshKDTree2D
    _triangle_to_cell_map: NDArray[int32]
    _grid_data: NDArray[float64]
    _fill_value: float

    def __init__(
        self,
        vertex_coords: ArrayLike,
        triangles: ArrayLike,
        triangle_to_cell_map: ArrayLike,
        grid_data: NDArray[float64],
        fill_value: float = 0.0,
    ) -> None: ...
    @classmethod
    def instance(
        cls,
        instance: UnstructGridFunction2D | UnstructGridVectorFunction2D,
        grid_data: NDArray[float64] | None = None,
        fill_value: float | None = None,
    ) -> UnstructGridFunction2D:
        """Create a new interpolator instance from an existing `UnstructGridFunction2D` or `UnstructGridVectorFunction2D` instance.

        The new interpolator instance will share the same internal acceleration data as the original
        interpolator. The grid_data of the new instance can be redefined.
        This method should be used if the user has multiple datasets that lie on the same mesh
        geometry. Using this methods avoids the repeated rebuilding of the mesh acceleration
        structures by sharing the geometry data between multiple interpolator objects.

        If created from the UnstructGridVectorFunction2D instance, the grid_data and the fill_value
        must not be None.

        Parameters
        ----------
        instance : UnstructGridFunction2D | UnstructGridVectorFunction2D
            The instance from which to create the new interpolator.
        grid_data : (L,) ndarray, optional
            Array containing data in the grid cells.
        fill_value : float, optional
            Value returned outside the grid, by default None.
            If None, inherited from the original instance.

        Returns
        -------
        UnstructGridFunction2D
            New interpolator instance.
        """

class UnstructGridVectorFunction2D(VectorFunction3D):
    """Simple vector interpolator for the data defined on the 2D unstructured grid.

    Find the cell containing the point (x, y) using the KDtree algorithm.
    Return the 3D vector value for this cell or the `fill_vector` if the grid does not contain the
    point.

    Parameters
    ----------
    vertex_coords : (N,3) array_like
        2D array-like with the vertex coordinates of triangles.
    triangles : (M,3) array_like
        2D integer array-like with the vertex indices forming the triangles.
    triangle_to_cell_map : (M,1) array_like
        1D integer array-like with the indices of the grid cells (polygons) containing the
        triangles.
    grid_vectors : (3,K) ndarray
        Array containing 3D vectors in the grid cells.
    fill_vector : Vector3D
        3D vector returned outside the gird, by default `Vector3D(0, 0, 0)`.
    """

    _kdtree: MeshKDTree2D
    _grid_vectors: NDArray[float64]
    _triangle_to_cell_map: NDArray[int32]
    _fill_vector: Vector3D

    def __init__(
        self,
        vertex_coords: ArrayLike,
        triangles: ArrayLike,
        triangle_to_cell_map: ArrayLike,
        grid_vectors: NDArray[float64],
        fill_vector: Vector3D = ZERO_VECTOR,
    ) -> None: ...
    @classmethod
    def instance(
        cls,
        instance: UnstructGridVectorFunction2D | UnstructGridFunction2D,
        grid_vectors: NDArray[float64] | None = None,
        fill_vector: Vector3D | None = None,
    ) -> UnstructGridVectorFunction2D:
        """Create a new interpolator instance from an existing `UnstructGridVectorFunction2D` or `UnstructGridFunction2D` instance.

        The new interpolator instance will share the same internal acceleration
        data as the original interpolator. The grid_vectors of the new instance can
        be redefined.
        This method should be used if the user has multiple datasets
        that lie on the same mesh geometry. Using this methods avoids the
        repeated rebuilding of the mesh acceleration structures by sharing the
        geometry data between multiple interpolator objects.

        If created from the UnstructGridFunction2D instance,
        the grid_vectors and the fill_vector must not be None.

        Parameters
        ----------
        instance : UnstructGridVectorFunction2D | UnstructGridFunction2D
            The instance from which to create the new interpolator.
        grid_vectors : (3,L) ndarray, optional
            Array containing vector grid data.
        fill_vector : Vector3D, optional
            3D vector returned outside the grid, by default None.
            If None, inherited from the original instance.

        Returns
        -------
        UnstructGridVectorFunction2D
            New interpolator instance.
        """



# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Module for loading core plasma profiles from the core_profiles IDS."""

from collections.abc import Callable
from dataclasses import fields

import numpy as np
from numpy.typing import NDArray
from raysect.core.math import translate
from raysect.core.math.function.float import Interpolator1DArray
from raysect.core.math.function.vector3d import Function2D as VectorFunction2D
from raysect.core.scenegraph._nodebase import _NodeBase
from raysect.primitive import Cylinder, Subtract
from scipy.constants import atomic_mass, electron_mass

from cherab.core import AtomicData, Maxwellian, Plasma, Species
from cherab.core.math import VectorAxisymmetricMapper
from cherab.imas.ids.core_profiles.load_profiles import ProfileData
from cherab.tools.equilibrium import EFITEquilibrium
from imas import DBEntry

from ..ids.common import get_ids_time_slice
from ..ids.core_profiles import load_core_grid, load_core_species
from .equilibrium import load_equilibrium, load_magnetic_field
from .utility import ZERO_VELOCITY, ProfileInterpolator, warn_unsupported_species

__all__ = ["load_core_plasma"]


def load_core_plasma(
    *args,
    time: float = 0,
    occurrence: int = 0,
    equilibrium: EFITEquilibrium | None = None,
    b_field: VectorFunction2D | None = None,
    psi_interpolator: Callable[[float], float] | None = None,
    time_threshold: float = np.inf,
    split_ion_bundles: bool = True,
    atomic_data: AtomicData | None = None,
    parent: _NodeBase | None = None,
    **kwargs,
) -> Plasma:
    """Load core profiles and create a `~cherab.core.plasma.node.Plasma` object.

    Prefer ``density_thermal`` over ``density`` profile.

    The distribution of each species is defined with `~cherab.core.distribution.Maxwellian` using
    its density and temperature profiles, which are mapped to 3D using the provided `equilibrium`.

    The plasma geometry is defined as a cylindrical annulus between the inner and outer radii of
    the equilibrium, and between the minimum and maximum z values of the equilibrium.

    The ion bundle species are split into their constituent charge states
    using `.solve_coronal_equilibrium` when `.split_ion_bundles` is True and the necessary atomic
    data is available. Otherwise, ion bundles are ignored with a warning.

    Parameters
    ----------
    *args
        Arguments passed to the `~imas.db_entry.DBEntry` constructor.
    time
        Time for the core plasma, by default 0.
    occurrence
        Occurrence index of the ``core_profiles`` IDS, by default 0.
    equilibrium
        Alternative `~cherab.tools.equilibrium.efit.EFITEquilibrium` used to map core
        profiles. By default None: the equilibrium is read from the same IMAS query as the
        core profiles. Ignored if the core plasma is not available.
    b_field
        Alternative 2D interpolator of the magnetic field vector (Br, Bphi, Bz).
        By default None: the magnetic field is loaded from the ``equilibrium`` IDS.
    psi_interpolator
        Alternative ``psi_norm(rho_tor_norm)`` interpolator.
        Used only if ``psi`` is missing in the core grid. By default None.
        Obtained from the ``equilibrium`` IDS.
    mask
        Mask function used for blending: ``(1 - mask) * f_edge + mask * f_core``.
        By default, uses `~cherab.tools.equilibrium.efit.EFITEquilibrium`'s `inside_lcfs`.
    time_threshold
        Maximum allowed difference between the requested time and the nearest
        available time, by default `numpy.inf`.
    split_ion_bundles
        Whether to split ion bundles into their constituent charge states using
        `.solve_coronal_equilibrium`, by default True.
    atomic_data
        Atomic data provider class for this plasma, by default None.
        If None, some species (e.g. ion_bundle) may not be properly loaded.
    parent
        Parent node in the Raysect scene graph, by default None.
        Typically a `~raysect.optical.scenegraph.world.World` instance.
    **kwargs
        Keyword arguments passed to the `~imas.db_entry.DBEntry` constructor.

    Returns
    -------
    `~cherab.core.plasma.node.Plasma`
        Plasma object with core profiles.

    Raises
    ------
    RuntimeError
        If the ``profiles_1d`` AOS in core_profiles IDS is empty.
    ValueError
        If the ``equilibrium`` argument is not an `~cherab.tools.equilibrium.efit.EFITEquilibrium`
        instance when provided.
    """
    # ---------------------------------------------------------
    # === Load core profiles and equilibrium data from IMAS ===
    # ---------------------------------------------------------
    # Load required data from the core_profiles IDS and form the core grid and species composition
    # data structures.
    with DBEntry(*args, **kwargs) as entry:
        core_profiles_ids = get_ids_time_slice(
            entry, "core_profiles", time=time, occurrence=occurrence, time_threshold=time_threshold
        )

    if not len(core_profiles_ids.profiles_1d):
        raise RuntimeError("The profiles_1d AOS in core_profiles IDS is empty.")

    # Load equilibrium and magnetic field data
    if equilibrium is None:
        equilibrium, psi_interp = load_equilibrium(
            *args, time=time, occurrence=occurrence, with_psi_interpolator=True, **kwargs
        )
        psi_interpolator = psi_interpolator or psi_interp

    if not isinstance(equilibrium, EFITEquilibrium):
        raise ValueError("Argument equilibrium must be a EFITEquilibrium instance.")

    if b_field is None:
        try:
            b_field = load_magnetic_field(*args, time=time, occurrence=occurrence, **kwargs)
        except RuntimeError:
            b_field = equilibrium.b_field

    # Create core grid
    core_grid = load_core_grid(core_profiles_ids.profiles_1d[0].grid)

    psi_norm = get_psi_norm(
        core_grid.psi,
        equilibrium.psi_axis,
        equilibrium.psi_lcfs,
        core_grid.rho_tor_norm,
        psi_interpolator,
    )

    # Load species composition
    composition = load_core_species(
        core_profiles_ids.profiles_1d[0],
        split_ion_bundles=split_ion_bundles,
        atomic_data=atomic_data,
    )

    # ----------------------------
    # === Create Plasma object ===
    # ----------------------------
    name = f"IMAS core plasma: time {core_profiles_ids.time[0]}, uri {entry.uri}."
    plasma = Plasma(parent=parent, name=name)
    radius_inner, radius_outer = equilibrium.r_range
    zmin, zmax = equilibrium.z_range
    height = zmax - zmin
    plasma.geometry = Subtract(Cylinder(radius_outer, height), Cylinder(radius_inner, height))
    plasma.geometry_transform = translate(0, 0, zmin)

    # Add magnetic field
    plasma.b_field = VectorAxisymmetricMapper(b_field)

    # Add atomic data
    if atomic_data is not None:
        plasma.atomic_data = atomic_data

    # ------------------------------------
    # === Define Electron Distribution ===
    # ------------------------------------
    if composition.electron.density_thermal is not None:
        composition.electron.density = composition.electron.density_thermal

    if composition.electron is None:
        raise RuntimeError("Electron profile is missing in the core_profiles IDS.")
    if composition.electron.density is None:
        raise RuntimeError("Electron density profile is missing in the core_profiles IDS.")
    if composition.electron.temperature is None:
        raise RuntimeError("Electron temperature profile is missing in the core_profiles IDS.")

    interp = get_core_interpolators(psi_norm, composition.electron, equilibrium, return3d=True)

    plasma.electron_distribution = Maxwellian(
        interp.density,
        interp.temperature,
        interp.velocity or ZERO_VELOCITY,
        electron_mass,
    )

    # -----------------------------------------------
    # === Define Species Distribution/Composition ===
    # -----------------------------------------------

    # === Ion/Neutral Species ===
    for profile in composition.ion + composition.neutral:
        if profile.species is None:
            print(f"Warning! Skipping species with missing element or charge: {profile}")
            continue
        if profile.species.element is None:
            print(f"Warning! Skipping species with missing element: {profile}")
            continue
        if profile.density_thermal is not None:
            profile.density = profile.density_thermal
        if profile.density is None:
            print(f"Warning! Skipping {profile.species}: density profile is missing.")
            continue
        if profile.temperature is None:
            print(f"Warning! Skipping {profile.species}: temperature profile is missing.")
            continue

        element = profile.species.element
        charge = profile.species.z_min

        try:
            species = plasma.composition.get(element, int(charge))
            print(f"Warning! Skipping {species}: already defined")
            continue
        except ValueError:
            pass

        interp = get_core_interpolators(psi_norm, profile, equilibrium, return3d=True)

        distribution = Maxwellian(
            interp.density,
            interp.temperature,
            interp.velocity or ZERO_VELOCITY,
            element.atomic_weight * atomic_mass,
        )

        plasma.composition.add(Species(element, int(charge), distribution))

    # === Ion Bundles ===
    # Ion bundles are split into their constituent charge states at the composition level.
    warn_unsupported_species(composition, "ion_bundle")

    # === Molecular Species ===
    # TODO: properly support molecular species.
    # For now, just issue a warning if any molecule or molecular_bundle species are present in the composition.
    warn_unsupported_species(composition, "molecule")
    warn_unsupported_species(composition, "molecular_bundle")

    return plasma


def get_core_interpolators(
    psi_norm: NDArray[np.float64],
    profile: ProfileData,
    equilibrium: EFITEquilibrium,
    return3d: bool = False,
) -> ProfileInterpolator:
    """Create interpolators for the core profiles.

    Parameters
    ----------
    psi_norm
        Normalized poloidal flux values.
    profile
        Instance of the `ProfileData` dataclass containing core plasma profiles.
    equilibrium
        `EFITEquilibrium` object used to map core profiles.
    return3d
        If True, return the 3D interpolators assuming rotational symmetry, by default False.

    Returns
    -------
    `.ProfileInterpolator`
        Instance of the `ProfileInterpolator` dataclass containing the interpolators for density and temperature.

    Raises
    ------
    ValueError
        If the ``equilibrium`` argument is not an `~cherab.tools.equilibrium.efit.EFITEquilibrium`
        instance.
    """
    if not isinstance(equilibrium, EFITEquilibrium):
        raise ValueError("Argument equilibrium must be a EFITEquilibrium instance.")

    psi_norm, index = np.unique(psi_norm, return_index=True)

    interpolators = ProfileInterpolator()

    for field in fields(profile):
        if field.name in {"species", "velocity"}:
            continue
        data_1d = getattr(profile, field.name, None)
        if isinstance(data_1d, np.ndarray) and data_1d.size > 0:
            extrapolation_range = max(0, psi_norm[0], 1.0 - psi_norm[-1])
            func = Interpolator1DArray(
                psi_norm, data_1d[index], "cubic", "nearest", extrapolation_range
            )
            setattr(
                interpolators,
                field.name,
                equilibrium.map3d(func) if return3d else equilibrium.map2d(func),
            )

    if profile.velocity is not None:
        pass  # TODO: handle velocity profile

    return interpolators


def get_psi_norm(
    psi: NDArray[np.float64] | None,
    psi_axis: float,
    psi_lcfs: float,
    rho_tor_norm: NDArray[np.float64] | None,
    psi_interpolator: Callable[[float], float] | None,
) -> NDArray[np.float64]:
    """Calculate normalized poloidal flux.

    Parameters
    ----------
    psi
        Poloidal flux values from the core grid.
    psi_axis
        Poloidal flux at the magnetic axis.
    psi_lcfs
        Poloidal flux at the last closed flux surface.
    rho_tor_norm
        Normalized toroidal flux values.
    psi_interpolator
        Interpolator function to map `rho_tor_norm` to `psi_norm`.
        Used only if ``psi`` is None.

    Returns
    -------
    `NDArray[np.float64]`
        Normalized poloidal flux values.

    Raises
    ------
    RuntimeError
        If both ``psi`` and ``rho_tor_norm`` are None, or if ``psi_interpolator`` is None when
        ``psi`` is None.
    """
    if psi is None:
        if psi_interpolator is None:
            raise RuntimeError(
                "Unable to map rho_tor_norm to psi_norm grid: psi_interpolator is not provided."
            )

        if rho_tor_norm is None:
            raise RuntimeError(
                "No rho_tor_norm values are available in the core grid: unable to interpolate to psi_norm."
            )

        return np.array([psi_interpolator(rho) for rho in rho_tor_norm])

    return (-psi / (2 * np.pi) - psi_axis) / (psi_lcfs - psi_axis)



"""Module for tetrahedralization of hexahedral cells."""
import numpy as np

cimport cython
from numpy cimport import_array, ndarray, int32_t

from cython.parallel import prange

__all__ = ["cell_to_5tetra", "cell_to_6tetra", "calculate_tetra_volume"]


import_array()


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cpdef ndarray[int32_t, ndim=2] cell_to_5tetra(const int32_t[:, ::1] cells):
    """Generate tetrahedral indices by dividing one cell into 5 tetrahedra.

    One cubic-like cell having 8 vertices can be divided into a minimum of five tetrahedra.
    Reference: https://www.mathworks.com/matlabcentral/mlc-downloads/downloads/submissions/48509/versions/3/previews/COMP_GEOM_TLBX/html/Divide_hypercube_5_simplices_3D.html  # noqa: E501

    .. note::
        If one side face of the cell is twisted, Adjacent cells do not share a plane with each
        other.
        In this case, it is required to alternate combinations of tetrahedra, or simply use
        :func:`~cherab.imas.math.tetrahedralize.cell_to_6tetra`.

    Parameters
    ----------
    cells : (N,8) ndarray [numpy.int32]
        cell indices 2D array, the shape of which is :math:`(N, 8)`, where :math:`N` is the number
        of cells.

    Returns
    -------
    `(5N,4) ndarray`
        tetrahedra indices array, the shape of which is :math:`(5N, 4)`.

    Examples
    --------
    >>> import numpy as np
    >>> from cherab.imas.math.tetrahedralize import cell_to_5tetra
    >>>
    >>> array = np.arange(16, dtype=np.int32).reshape((2, -1))
    >>> array
    array([[ 0,  1,  2,  3,  4,  5,  6,  7],
           [ 8,  9, 10, 11, 12, 13, 14, 15]], dtype=int32)
    >>> cell_to_5tetra(array)
        array([[ 0,  1,  3,  4],
               [ 1,  3,  4,  6],
               [ 3,  6,  7,  4],
               [ 1,  2,  3,  6],
               [ 1,  6,  4,  5],
               [ 8,  9, 11, 12],
               [ 9, 11, 12, 14],
               [11, 14, 15, 12],
               [ 9, 10, 11, 14],
               [ 9, 14, 12, 13]], dtype=int32)
    """
    cdef:
        Py_ssize_t i, j, k
        int[5][4] tetra_indices
        ndarray[int32_t, ndim=2] tetrahedra
        int32_t[:, ::1] tetrahedra_mv

    if cells.ndim != 2:
        raise ValueError("cells must be a 2 dimensional array.")

    if cells.shape[1] != 8:
        raise ValueError("cells must have a shape of (N, 8).")

    # tetrahedra indices array
    tetrahedra = np.empty((cells.shape[0] * 5, 4), dtype=np.int32)

    # five tetrahedra indices at one cell
    tetra_indices[0][:] = [0, 1, 3, 4]
    tetra_indices[1][:] = [1, 3, 4, 6]
    tetra_indices[2][:] = [3, 6, 7, 4]
    tetra_indices[3][:] = [1, 2, 3, 6]
    tetra_indices[4][:] = [1, 6, 4, 5]

    # memory view
    tetrahedra_mv = tetrahedra

    for i in prange(cells.shape[0], nogil=True):
        for j in range(5):
            for k in range(4):
                tetrahedra_mv[5 * i + j, k] = cells[i, tetra_indices[j][k]]

    return tetrahedra


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cpdef ndarray[int32_t, ndim=2] cell_to_6tetra(const int32_t[:, ::1] cells):
    """Generate tetrahedral indices by dividing one cell into 6 tetrahedra.

    One cubic-like cell having 8 vertices can be divided into six tetrahedra.
    This manner is useful when the cell is twisted.

    Parameters
    ----------
    cells : (N,8) ndarray [numpy.int32]
        Cell indices 2D array, the shape of which is :math:`(N, 8)`, where :math:`N` is the number
        of cells.

    Returns
    -------
    `(6N,4) ndarray`
        Tetrahedra indices array, the shape of which is :math:`(6N, 4)`.

    Examples
    --------
    >>> import numpy as np
    >>> from cherab.imas.math.tetrahedralize import cell_to_6tetra
    >>>
    >>> array = np.arange(16, dtype=np.int32).reshape((2, -1))
    >>> array
    array([[ 0,  1,  2,  3,  4,  5,  6,  7],
           [ 8,  9, 10, 11, 12, 13, 14, 15]], dtype=int32)
    >>> cell_to_6tetra(array)
    array([[ 6,  2,  1,  0],
           [ 7,  3,  2,  0],
           [ 0,  7,  6,  2],
           [ 1,  5,  6,  4],
           [ 0,  4,  6,  7],
           [ 6,  4,  0,  1],
           [14, 10,  9,  8],
           [15, 11, 10,  8],
           [ 8, 15, 14, 10],
           [ 9, 13, 14, 12],
           [ 8, 12, 14, 15],
           [14, 12,  8,  9]], dtype=int32)
    """
    cdef:
        int i, j, k
        int[6][4] tetra_indices
        ndarray[int32_t, ndim=2] tetrahedra
        int32_t[:, ::1] tetrahedra_mv

    if cells.ndim != 2:
        raise ValueError("cells must be a 2 dimensional array.")

    if cells.shape[1] != 8:
        raise ValueError("cells must have a shape of (N, 8).")

    # tetrahedra indices array
    tetrahedra = np.empty((cells.shape[0] * 6, 4), dtype=np.int32)

    # six tetrahedra indices at one cell
    tetra_indices[0][:] = [6, 2, 1, 0]
    tetra_indices[1][:] = [7, 3, 2, 0]
    tetra_indices[2][:] = [0, 7, 6, 2]
    tetra_indices[3][:] = [1, 5, 6, 4]
    tetra_indices[4][:] = [0, 4, 6, 7]
    tetra_indices[5][:] = [6, 4, 0, 1]

    # memory view
    tetrahedra_mv = tetrahedra

    for i in prange(cells.shape[0], nogil=True):
        for j in range(6):
            for k in range(4):
                tetrahedra_mv[6 * i + j, k] = cells[i, tetra_indices[j][k]]

    return tetrahedra


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cpdef double calculate_tetra_volume(
    const double[:, ::1] vertices,
    const int32_t[:, ::1] tets,
) nogil:
    """Calculate the volume of tetrahedra.

    Parameters
    ----------
    vertices : ndarray [numpy.float64]
        Vertices of tetrahedra.
    tets : ndarray [numpy.int32]
        Tetrahedra indices.

    Returns
    -------
    float
        Volume of tetrahedra.

    Examples
    --------
    >>> vertices = np.array([[0.0, 0.0, 0.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]])
    >>> tetras = np.array([[0, 1, 2, 3]], dtype=np.int32)
    >>> calculate_tetra_volume(vertices, tetras)
    0.16666666666666666
    """
    cdef:
        double volume = 0.0
        int i
        double v0x, v0y, v0z, v1x, v1y, v1z, v2x, v2y, v2z, v3x, v3y, v3z
        double cx, cy, cz, dot_product

    for i in range(tets.shape[0]):
        v0x, v0y, v0z = vertices[tets[i, 0], 0], vertices[tets[i, 0], 1], vertices[tets[i, 0], 2]
        v1x, v1y, v1z = vertices[tets[i, 1], 0], vertices[tets[i, 1], 1], vertices[tets[i, 1], 2]
        v2x, v2y, v2z = vertices[tets[i, 2], 0], vertices[tets[i, 2], 1], vertices[tets[i, 2], 2]
        v3x, v3y, v3z = vertices[tets[i, 3], 0], vertices[tets[i, 3], 1], vertices[tets[i, 3], 2]

        cx = (v1y - v0y) * (v2z - v0z) - (v1z - v0z) * (v2y - v0y)
        cy = (v1z - v0z) * (v2x - v0x) - (v1x - v0x) * (v2z - v0z)
        cz = (v1x - v0x) * (v2y - v0y) - (v1y - v0y) * (v2x - v0x)

        dot_product = cx * (v3x - v0x) + cy * (v3y - v0y) + cz * (v3z - v0z)
        volume += (1.0 / 6.0) * abs(dot_product)

    return volume



# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Module for common functions used to get IDS time slices."""

import warnings

from numpy import inf

from imas.db_entry import DBEntry
from imas.ids_defs import CLOSEST_INTERP
from imas.ids_toplevel import IDSToplevel

__all__ = ["get_ids_time_slice"]


def get_ids_time_slice(
    entry: DBEntry,
    ids_name: str,
    time: float = 0,
    occurrence: int = 0,
    time_threshold: float = inf,
) -> IDSToplevel:
    """Get a time slice of the specified IDS from the given IMAS entry.

    .. note::
        If the `~imas.db_entry.DBEntry.get_slice` method is not implemented for the given IMAS entry
        URI, this function will fall back to using the `~imas.db_entry.DBEntry.get` method and
        return the entire IDS.

    Parameters
    ----------
    entry
        The IMAS entry. The entry must be opened in read mode.
    ids_name
        The name of the IDS.
    time
        The time in seconds of the requested time slice, by default is 0.
    occurrence
        The occurrence of the IDS, by default is 0.
    time_threshold
        The maximum allowed time difference in seconds between the actual time of the nearest time
        slice and the given time, by default is infinity.

    Returns
    -------
    `~imas.ids_toplevel.IDSToplevel`
        The requested IDS time slice.

    Raises
    ------
    ValueError
        If `.time` or `.time_threshold` is negative.
    RuntimeError
        If the requested IDS is empty.
    RuntimeError
        If the time difference between the actual time of the nearest time slice and the given time
        exceeds the specified threshold.

    Examples
    --------
    .. code-block:: python

        from imas import DBEntry
        from cherab.imas.ids.common import get_ids_time_slice

        with DBEntry(
            "imas://uda.iter.org/uda?path=/work/imas/shared/imasdb/ITER/3/123072/3&backend=hdf5",
            "r",
        ) as entry:
            ids = get_ids_time_slice(entry, "equilibrium", time=0.0)
    """
    if time < 0:
        raise ValueError(f"Argument 'time' must be >=0 ({time} s).")
    if time_threshold < 0:
        raise ValueError(f"Argument 'time_threshold' must be >=0 ({time_threshold} s).")

    try:
        ids = entry.get_slice(
            ids_name,
            time,
            CLOSEST_INTERP,
            occurrence=occurrence,
            autoconvert=False,
        )
    except NotImplementedError:
        # Fallback to `get` method to retrieve the entire IDS
        warnings.warn(
            f"The 'get_slice' method is not implemented for the URI '{entry.uri}'. "
            + "Falling back to 'get' method to retrieve the entire IDS.",
            RuntimeWarning,
            stacklevel=2,
        )
        ids = entry.get(ids_name, occurrence=occurrence, autoconvert=False)

    if not len(ids.time):
        raise RuntimeError(f"The '{ids_name}' IDS is empty.")

    if abs(ids.time[0] - time) > time_threshold:
        raise RuntimeError(
            f"The time difference between the actual time ({ids.time[0]} s) "
            + f"of the nearest '{ids_name}' time slice and the given time ({time} s) "
            + f"exceeds the specified threshold ({time_threshold} s)."
        )

    return ids



"""Module defining simple interpolators for data defined on a 3D structured grid."""

from numpy import float64
from numpy.typing import ArrayLike, NDArray
from raysect.core.math import Vector3D
from raysect.core.math.function.float import Function3D
from raysect.core.math.function.vector3d import Function3D as VectorFunction3D

ZERO_VECTOR = Vector3D(0, 0, 0)

class StructGridFunction3D(Function3D):
    """Simple interpolator for the data defined on the 3D structured grid.

    Find the cell containing the point (x, y, z).
    Return the data value for this cell or the `fill_value` if the points lies outside the grid.

    Parameters
    ----------
    x : (L,) array_like
        The corners of the quadrilateral cells along x axis.
    y : (M,) array_like
        The corners of the quadrilateral cells along y axis.
    z : (N,) array_like
        The corners of the quadrilateral cells along z axis.
    grid_data : (L-1, M-1, N-1) ndarray
        Array containing data in the grid cells.
    fill_value : float, optional
        A value returned outside the grid, by default 0.
    """

    _x: NDArray[float64]
    _y: NDArray[float64]
    _z: NDArray[float64]
    _grid_data: NDArray[float64]
    _fill_value: float
    def __init__(
        self,
        x: ArrayLike,
        y: ArrayLike,
        z: ArrayLike,
        grid_data: NDArray[float64],
        fill_value: float = 0.0,
    ) -> None: ...

class StructGridVectorFunction3D(VectorFunction3D):
    """Simple vector interpolator for the data defined on the 3D structured grid.

    Find the cell containing the point (x, y, z).
    Return the 3D vector value this cell or the `fill_vector` if the points lies outside the grid.

    Parameters
    ----------
    x : (L,) array_like
        The corners of the quadrilateral cells along x axis.
    y : (M,) array_like
        The corners of the quadrilateral cells along y axis.
    z : (N,) array_like
        The corners of the quadrilateral cells along z axis.
    grid_vectors : (3, L-1, M-1, N-1) ndarray
        Array containing 3D vectors in the grid cells.
    fill_vector : Vector3D
        3D vector returned outside the gird, by default `Vector3D(0, 0, 0)`.
    """

    _x: NDArray[float64]
    _y: NDArray[float64]
    _z: NDArray[float64]
    _grid_vectors: NDArray[float64]
    _fill_vector: Vector3D
    def __init__(
        self,
        x: ArrayLike,
        y: ArrayLike,
        z: ArrayLike,
        grid_vectors: NDArray[float64],
        fill_vector: Vector3D = ZERO_VECTOR,
    ) -> None: ...



# cython: language_level=3

# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

cimport numpy as np
from raysect.core.math.vector cimport Vector3D
from raysect.core.math.function.float.function2d.interpolate.common cimport MeshKDTree2D
from raysect.core.math.function.float.function2d cimport Function2D
from raysect.core.math.function.vector3d.function2d cimport Function2D as VectorFunction2D


cdef class UnstructGridFunction2D(Function2D):

    cdef:
        MeshKDTree2D _kdtree
        np.ndarray _grid_data, _triangle_to_cell_map
        np.int32_t[::1] _triangle_to_cell_map_mv
        double[::1] _grid_data_mv
        double _fill_value

cdef class UnstructGridVectorFunction2D(VectorFunction2D):

    cdef:
        MeshKDTree2D _kdtree
        np.ndarray _grid_vectors, _triangle_to_cell_map
        np.int32_t[::1] _triangle_to_cell_map_mv
        double[:, ::1] _grid_vectors_mv
        Vector3D _fill_vector



"""Module defining unit vector functions."""

from collections.abc import Callable

from raysect.core.math import Vector3D
from raysect.core.math.function.vector3d import Function1D as VectorFunction1D
from raysect.core.math.function.vector3d import Function2D as VectorFunction2D
from raysect.core.math.function.vector3d import Function3D as VectorFunction3D

__all__ = ["UnitVector1D", "UnitVector2D", "UnitVector3D"]

class UnitVector1D(VectorFunction1D):
    """Evaluate a unit vector for the given VectorFunction1D instance."""

    _vector: VectorFunction1D
    def __init__(
        self, vector: VectorFunction1D | tuple[float, float, float] | Callable[[float], Vector3D]
    ) -> None: ...

class UnitVector2D(VectorFunction2D):
    """Evaluate a unit vector for the given VectorFunction2D instance."""

    _vector: VectorFunction2D
    def __init__(
        self,
        vector: VectorFunction2D | tuple[float, float, float] | Callable[[float, float], Vector3D],
    ) -> None: ...

class UnitVector3D(VectorFunction3D):
    """Evaluate a unit vector for the given VectorFunction3D instance."""

    _vector: VectorFunction3D
    def __init__(
        self,
        vector: VectorFunction3D
        | tuple[float, float, float]
        | Callable[[float, float, float], Vector3D],
    ) -> None: ...



from plotly import graph_objs as go

from cherab.imas.observer.bolometer import load_bolometers, visualize


def test_load_bolometers(path_bolometer_moc: str) -> None:
    """Test loading bolometer data from an IDS dataset."""
    bolometers = load_bolometers(path_bolometer_moc, "r")

    # Check that the bolometer cameras are loaded correctly
    assert len(bolometers) == 3

    # Check the visualization (this is a smoke test to ensure it runs without errors)
    for bolo in bolometers:
        fig = visualize(
            bolo, num_rays=100, ray_from_channel=0, ray_terminate_distance=1e-2, show=False
        )
        assert isinstance(fig, go.Figure)

    fig = visualize(
        bolometers[-1],
        num_rays=100,
        ray_from_channel=[0, 3],
        ray_terminate_distance=1e-2,
        show=False,
    )
    assert isinstance(fig, go.Figure)



import pytest

from cherab.core.atomic.elements import neon
from cherab.openadas import OpenADAS
from cherab.openadas.repository import populate


@pytest.fixture(scope="session", autouse=True)
def populate_openadas_repository():
    """Fixture to populate the OpenADAS repository before running tests."""
    adas = OpenADAS(missing_rates_return_null=False)
    try:
        adas.ionisation_rate(neon, 1)
    except Exception:
        print("Populating OpenADAS repository...")
        populate()



# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Module for loading wall components from wall IDSs."""

import numpy as np
from raysect.core.scenegraph._nodebase import _NodeBase
from raysect.optical.material.material import Material
from raysect.primitive import Mesh

from imas import DBEntry

from ..ids.common import get_ids_time_slice
from ..ids.wall import load_wall_2d, load_wall_3d

__all__ = ["load_wall_mesh", "load_wall_outline"]


def load_wall_mesh(
    *args,
    time: float = 0,
    occurrence: int = 0,
    desc_ggd_index: int = 0,
    subsets: list[str] | None = None,
    materials: dict[str, Material] | None = None,
    time_threshold: float = np.inf,
    parent: _NodeBase | None = None,
    **kwargs,
) -> dict[str, Mesh]:
    """Load machine wall components from IMAS wall IDS and Create Raysect mesh primitives.

    Parameters
    ----------
    *args
        Arguments passed to the `~imas.db_entry.DBEntry` constructor.
    time
        Time for the wall, by default 0.
    occurrence
        Occurrence index of the ``wall`` IDS, by default 0.
    desc_ggd_index
        Index of ``description_ggd``, by default 0.
    subsets
        List of names of specific ggd subsets to load, by default None (loads all subsets).
    materials
        Optional dictionary with Raysect materials for each wall component, by default None.
        Use component names as keys. The components are split by their grid subsets and for
        each grid subset by materials. All elements of the grid subset that share the same material
        are combined into a single component. The component names are assigns as follows:
        ``"{grid_name}.{subset_name}.{material_name}"``
        E.g.: ``"TokamakWall.full_main_chamber_wall.Be"``.
    time_threshold
        Maximum allowed difference between the requested time and the nearest available
        time, by default `numpy.inf`.
    parent
        Parent node in the Raysect scene graph, by default None.
        Normally, `~raysect.optical.scenegraph.world.World` instance.
    **kwargs
        Keyword arguments passed to the `~imas.db_entry.DBEntry` constructor.

    Returns
    -------
    dict[str, Mesh]
        Dictionary with the Raysect Mesh instances.

    Examples
    --------
    >>> from raysect.optical import World
    >>> world = World()
    >>> meshes = load_wall_mesh(
    ...     "imas:hdf5?path=/work/imas/shared/imasdb/ITER_MD/3/116100/1001/", "r", parent=world
    ... )
    >>> meshes
    {'FullTokamak.none.none': <raysect.primitive.mesh.mesh.Mesh at 0x1766322a0>}
    """
    with DBEntry(*args, **kwargs) as entry:
        wall_ids = get_ids_time_slice(
            entry, "wall", time=time, occurrence=occurrence, time_threshold=time_threshold
        )

    wall_dict = load_wall_3d(wall_ids.description_ggd[desc_ggd_index], subsets)

    components = {}

    for key, value in wall_dict.items():
        mesh = Mesh(value["vertices"], value["triangles"], closed=False)
        mesh.parent = parent
        mesh.name = key
        components[key] = mesh
        if materials:
            mesh.material = materials[key]

    return components


def load_wall_outline(
    *args, occurrence: int = 0, desc_index: int = 0, **kwargs
) -> dict[str, np.ndarray]:
    """Load 2D wall outline (limiter contour only) from IMAS wall IDS.

    Parameters
    ----------
    *args
        Arguments passed to the `~imas.db_entry.DBEntry` constructor.
    occurrence
        Occurrence index of the ``wall`` IDS, by default 0.
    desc_index
        Index of ``description_2d``, by default 0.
    **kwargs
        Keyword arguments passed to the `~imas.db_entry.DBEntry` constructor.

    Returns
    -------
    dict[str, (N, 2) ndarray]
       Dictionary of wall unit outlines ``(N, 2)`` array given in RZ coordinates.

    Examples
    --------
    >>> load_wall_outline("imas:hdf5?path=/work/imas/shared/imasdb/ITER_MD/3/116000/5/", "r")
    {'First Wall': array([[ 4.11129713, -2.49559808],
                          [ 4.11129713, -1.48329401],
                          ...
                          [ 6.21105623, -3.06856108]]),
     'Divertor': array([[ 3.94210005, -2.53570008],
                        ...
                        [ 6.36320019, -3.24460006]])}
    """
    with DBEntry(*args, **kwargs) as entry:
        description2d = entry.get("wall", occurrence=occurrence, autoconvert=False).description_2d[
            desc_index
        ]

    wall_outline = load_wall_2d(description2d)

    return wall_outline



# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Module for loading 2D wall description from IMAS wall IDS."""

import numpy as np

from imas.ids_structure import IDSStructure

__all__ = ["load_wall_2d"]


def load_wall_2d(description_2d: IDSStructure) -> dict[str, np.ndarray]:
    """Load 2D wall outline (limiter contour only) from IMAS wall IDS.

    Parameters
    ----------
    description_2d
        IDS structure with 2D description of the wall.

    Returns
    -------
    `dict[str, numpy.ndarray]`
        Dictionary of wall unit outlines given in RZ coordinates.
    """
    wall_outline = {}

    for unit in description_2d.limiter.unit:
        r = unit.outline.r
        z = unit.outline.z
        wall_outline[unit.name.value] = np.array([r, z]).T

    return wall_outline



import numpy as np
import pytest
from imas import DBEntry
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from raysect.core.scenegraph import World
from raysect.primitive import Subtract

from cherab.core import Maxwellian, Plasma
from cherab.core.atomic.elements import neon
from cherab.imas.ids.common import get_ids_time_slice
from cherab.imas.ids.common.ggd import load_grid
from cherab.imas.ids.edge_profiles import load_edge_species
from cherab.imas.plasma.edge import load_edge_plasma
from cherab.imas.plasma.equilibrium import load_magnetic_field

plt.rcParams["backend"] = "Agg"  # Use non-interactive backend for testing


def test_load_edge_plasma(path_iter_jintrac: str):
    """Test loading of edge plasma data from an IMAS file."""
    plasma = load_edge_plasma(path_iter_jintrac, "r")

    # Test that a Plasma object is returned
    assert isinstance(plasma, Plasma)

    # Check that geometry is set and has the expected type
    assert plasma.geometry is not None
    assert isinstance(plasma.geometry, Subtract)

    # Check that geometry transform is set
    assert plasma.geometry_transform is not None

    # Check that electron distribution exists
    assert plasma.electron_distribution is not None
    assert isinstance(plasma.electron_distribution, Maxwellian)

    # Check if a certain bundled species is split into individual ion species.
    # Here we check only the Ne species as an example, but this can be extended to other species
    # based on the test dataset.
    assert len(plasma.composition) > 0
    ion_charges = set()
    for species in plasma.composition:
        # Check that we have multiple ionization states for Ne
        if species.element == neon:
            ion_charges.add(species.charge)

    # Expecting all charge states from 0 to Z for Ne
    assert len(ion_charges) == neon.atomic_number + 1


def test_load_edge_plasma_with_time(path_iter_jintrac: str):
    """Test loading edge plasma with specific time parameter."""
    plasma = load_edge_plasma(path_iter_jintrac, "r", time=0.1, split_ion_bundles=False)
    assert isinstance(plasma, Plasma)

    with pytest.raises(ValueError):
        load_edge_plasma(path_iter_jintrac, "r", time=-1, split_ion_bundles=False)


def test_load_edge_plasma_with_grid_subset_id(path_iter_jintrac: str):
    """Test loading edge plasma with different grid subset IDs."""
    # Test with string identifier
    plasma = load_edge_plasma(
        path_iter_jintrac, "r", grid_subset_id="cells", split_ion_bundles=False
    )
    assert isinstance(plasma, Plasma)


def test_load_edge_plasma_with_parent(path_iter_jintrac: str):
    """Test loading edge plasma with a parent scene graph node."""
    world = World()

    plasma = load_edge_plasma(path_iter_jintrac, "r", parent=world, split_ion_bundles=False)
    assert isinstance(plasma, Plasma)
    assert plasma.parent is world


def test_load_edge_plasma_with_magnetic_field(path_iter_jintrac: str):
    """Test loading edge plasma with external magnetic field."""
    # Load magnetic field separately
    try:
        b_field = load_magnetic_field(path_iter_jintrac, "r")
        plasma = load_edge_plasma(path_iter_jintrac, "r", b_field=b_field, split_ion_bundles=False)
        assert isinstance(plasma, Plasma)
        assert plasma.b_field is not None
    except RuntimeError:
        # If magnetic field loading fails, test should still pass
        # as this is handled gracefully in the function
        pass


def test_load_edge_plasma_time_threshold(path_iter_jintrac: str):
    """Test loading edge plasma with time threshold parameter."""
    # Test with different time thresholds
    plasma = load_edge_plasma(
        path_iter_jintrac, "r", time_threshold=500, split_ion_bundles=False
    )  # test dataset has time ~450s only
    assert isinstance(plasma, Plasma)

    with pytest.raises(RuntimeError):
        load_edge_plasma(path_iter_jintrac, "r", time_threshold=0.0, split_ion_bundles=False)


def test_edge_plasma_profiles(path_iter_jintrac: str):
    """Test that edge plasma profiles can be loaded and plotted."""
    with DBEntry(path_iter_jintrac, "r") as entry:
        ids = get_ids_time_slice(entry, "edge_profiles", time=0)

    # Load mesh grid
    grid, subsets, subset_id = load_grid(ids.grid_ggd[0], with_subsets=True)
    grid_subset_name = "Cells"
    try:
        grid = grid.subset(subsets[grid_subset_name], name=grid_subset_name)
    except KeyError:
        grid_subset_name = grid_subset_name.lower()
        grid = grid.subset(subsets[grid_subset_name], name=grid_subset_name)

    # Load edge species composition
    composition_w_bundle = load_edge_species(
        ids.ggd[0],
        grid_subset_index=subset_id[grid_subset_name],
        split_ion_bundles=False,
    )
    composition = load_edge_species(
        ids.ggd[0],
        grid_subset_index=subset_id[grid_subset_name],
        split_ion_bundles=True,
    )

    # Check that the summed density of the split ion species matches the original bundled species density
    ion_bundles = []
    for bundle in composition_w_bundle.ion_bundle:
        if bundle.species.element == neon and bundle.density is not None:
            ion_bundles += [bundle]
    if not ion_bundles:
        raise ValueError("Test dataset does not contain a bundled Ne ion species with density.")
    for ion_bundle in ion_bundles:
        original_density = ion_bundle.density
        element = ion_bundle.species.element
        z_min = ion_bundle.species.z_min
        z_max = ion_bundle.species.z_max
        split_density_sum = np.zeros_like(original_density)
        for profile in composition.ion:
            if (
                profile.species.element == element
                and profile.species.z_min >= z_min
                and profile.species.z_max <= z_max
            ):
                if profile.density is None:
                    raise ValueError(
                        f"Test dataset does not contain density for split ion species: {profile.species}."
                    )
                split_density_sum += profile.density

        np.testing.assert_allclose(split_density_sum, original_density)

    # Plot electron density
    electron_density = composition.electron.density
    ax = grid.plot_mesh(data=electron_density)

    assert isinstance(ax, Axes)


def test_load_edge_plasma_not_split_ion_bundles(path_iter_jintrac: str):
    """Test that loading with split_ion_bundles=False."""
    plasma = load_edge_plasma(path_iter_jintrac, "r", split_ion_bundles=False)

    ion_charges_no_split = set()
    for species in plasma.composition:
        if species.element == neon:
            ion_charges_no_split.add(species.charge)

    # Expecting only the bundled charge state
    assert 0 < len(ion_charges_no_split) < neon.atomic_number + 1



import shutil
from pathlib import Path

import pytest

from cherab.imas.datasets import bolometer_moc


@pytest.fixture(scope="session")
def path_bolometer_moc(tmp_path_factory) -> str:
    """Fixture to provide the path to a sample bolometer IMAS dataset."""
    path = Path(bolometer_moc())
    tmp_path = tmp_path_factory.mktemp("cherab-imas-data")
    shutil.copy2(path, tmp_path)
    return str(tmp_path / path.name)



# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Interpolators for structured and unstructured grid functions.

This module provides interpolation classes for 2D and 3D grid-based functions, supporting both
structured and unstructured grids. These interpolators are designed to work with IMAS data
structures and provide efficient evaluation of scalar and vector functions defined on various grid
types.
"""

from .struct_grid_2d_functions import StructGridFunction2D, StructGridVectorFunction2D
from .struct_grid_3d_functions import StructGridFunction3D, StructGridVectorFunction3D
from .unstruct_grid_2d_functions import UnstructGridFunction2D, UnstructGridVectorFunction2D
from .unstruct_grid_3d_functions import UnstructGridFunction3D, UnstructGridVectorFunction3D

__all__ = [
    "StructGridFunction2D",
    "StructGridVectorFunction2D",
    "StructGridFunction3D",
    "StructGridVectorFunction3D",
    "UnstructGridFunction2D",
    "UnstructGridVectorFunction2D",
    "UnstructGridFunction3D",
    "UnstructGridVectorFunction3D",
]



# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Module for loading magnetic field data from the equilibrium IDS."""

from dataclasses import dataclass
from warnings import warn

import numpy as np

from imas.ids_defs import EMPTY_INT
from imas.ids_struct_array import IDSStructArray

RECTANGULAR_GRID = 1

__all__ = ["load_magnetic_field_data"]


@dataclass
class MagneticField2DData:
    """Dataclass to hold 2D magnetic field data."""

    r: np.ndarray
    z: np.ndarray
    b_field_r: np.ndarray
    b_field_z: np.ndarray
    b_field_phi: np.ndarray


def load_magnetic_field_data(profiles_2d: IDSStructArray) -> MagneticField2DData:
    """Load 2D profiles of the magnetic field components from equilibrium IDS.

    The magnetic field components are extracted from the ``profiles_2d`` IDS structure,
    assuming that the profiles are defined on a rectangular grid.

    Parameters
    ----------
    profiles_2d
        The ``profiles_2d`` structure of the equilibrium IDS.

    Returns
    -------
    MagneticField2DData
        Dataclass containing the magnetic field data:

        :r: ``(N,)`` ndarray with R coordinates of rectangular grid.
        :z: ``(M,)`` ndarray with Z coordinates of rectangular grid.
        :b_field_r: ``(N, M)`` ndarray with R component of the magnetic field.
        :b_field_z: ``(N, M)`` ndarray with Z component of the magnetic field.
        :b_field_phi: ``(N, M)`` ndarray with toroidal component of the magnetic field.

    Raises
    ------
    RuntimeError
        If unable to read the magnetic field due to unsupported grid type or
        mismatched array shapes.
    RuntimeError
        If the toroidal component of the magnetic field is not found in the IDS structure.
    """
    rectangular_grid = False
    for prof2d in profiles_2d:
        if prof2d.grid_type.index == RECTANGULAR_GRID or prof2d.grid_type.index == EMPTY_INT:
            rectangular_grid = True
            break

    if not rectangular_grid:
        raise RuntimeError(
            "Unable to read magnetic field: "
            + "rectangular grid for 2D profiles is not found and other grid types are not supported."
        )

    r = np.asarray_chkfinite(prof2d.grid.dim1)
    z = np.asarray_chkfinite(prof2d.grid.dim2)
    shape = (r.size, z.size)

    b_field_r = np.asarray_chkfinite(prof2d.b_field_r)
    b_field_z = np.asarray_chkfinite(prof2d.b_field_z)

    # TODO: Remove fallback to `b_field_tor` after confirming that all relevant IDS structures have `b_field_phi`
    if hasattr(prof2d, "b_field_phi"):
        b_field_phi = np.asarray_chkfinite(prof2d.b_field_phi)
    elif hasattr(prof2d, "b_field_tor"):
        warn(
            "The 'b_field_tor' field is deprecated since DD v3.42.0; "
            "Support for 'b_field_tor' fallback will be removed in a future release.",
            DeprecationWarning,
            stacklevel=2,
        )
        b_field_phi = np.asarray_chkfinite(prof2d.b_field_tor)
    else:
        raise RuntimeError(
            "Unable to read magnetic field: "
            + "toroidal component of the magnetic field is not found in the IDS structure."
        )

    if b_field_r.shape != shape or b_field_phi.shape != shape or b_field_z.shape != shape:
        raise RuntimeError(
            "Unable to read magnetic field: "
            + "the shape of the magnetic field components does not match the grid shape."
        )

    return MagneticField2DData(
        r=r,
        z=z,
        b_field_r=b_field_r,
        b_field_z=b_field_z,
        b_field_phi=b_field_phi,
    )



"""Serve as the IMAS sample dataset registry."""

registry = {
    "iter_disruption_113112_1.nc": "md5:f8747539b8d5a4b6974cc1bb33f2924a",  # JOREK 3D simulation
    "iter_scenario_123364_1.nc": "md5:b969c91e2c0f2df9a500edd55829764b",  # SOLPS-ITER simulation
    "iter_scenario_53298_seq1_DD4.nc": "md5:6bd52d3dd9b456e86aab6de5c0e0788a",  # JINTRAC simulation
}

# dataset method mapping with their associated filenames
# <method_name> : ["filename1", "filename2", ...]
method_files_map = {
    "iter_jintrac": ["iter_scenario_53298_seq1_DD4.nc", "iter_scenario_53298_seq1_DD4_mod.nc"],
    "iter_solps": ["iter_scenario_123364_1.nc"],
    "iter_jorek": ["iter_disruption_113112_1.nc"],
    "bolometer_moc": ["bolometer_moc.nc"],
}



import tempfile
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest

from cherab.imas.datasets import clear_cache, iter_jintrac, iter_solps
from cherab.imas.datasets._registry import method_files_map, registry
from cherab.imas.datasets._utils import _clear_cache


class TestClearCache:
    """Test suite for the clear_cache function."""

    def test_clear_cache_no_args_with_mock(self):
        """Test clear_cache with no arguments (should clear all cache)."""
        with patch("cherab.imas.datasets._utils._clear_cache") as mock_clear:
            clear_cache()
            mock_clear.assert_called_once_with(None)

    def test_clear_cache_single_dataset(self):
        """Test clear_cache with a single dataset function."""
        with patch("cherab.imas.datasets._utils._clear_cache") as mock_clear:
            clear_cache(iter_jintrac)
            mock_clear.assert_called_once_with(iter_jintrac)

    def test_clear_cache_multiple_datasets(self):
        """Test clear_cache with multiple dataset functions."""
        datasets = [iter_jintrac, iter_solps]
        with patch("cherab.imas.datasets._utils._clear_cache") as mock_clear:
            clear_cache(datasets)
            mock_clear.assert_called_once_with(datasets)

    def test_clear_cache_with_temp_directory(self):
        """Test _clear_cache with a temporary directory setup."""
        # Create a temporary directory structure
        with tempfile.TemporaryDirectory() as temp_dir:
            cache_dir = Path(temp_dir) / "test_cache"
            cache_dir.mkdir(parents=True)

            # Create test files matching the method_files_map
            test_files = list(registry.keys())

            for file in test_files:
                (cache_dir / file).touch()

            # Verify files exist before clearing
            assert all((cache_dir / file).exists() for file in test_files)

            # Clear all cache
            _clear_cache(None, cache_dir)

            # Verify cache directory is removed
            assert not cache_dir.exists()

    def test_clear_cache_single_dataset_with_temp_directory(self, capsys):
        """Test _clear_cache with a single dataset and temporary directory."""
        with tempfile.TemporaryDirectory() as temp_dir:
            cache_dir = Path(temp_dir) / "test_cache"
            cache_dir.mkdir(parents=True)

            # Create test files
            jintrac_file = method_files_map["iter_jintrac"][0]
            solps_file = method_files_map["iter_solps"][0]

            (cache_dir / jintrac_file).touch()
            (cache_dir / solps_file).touch()

            # Clear only jintrac dataset
            _clear_cache(iter_jintrac, cache_dir)

            # Verify only jintrac file is removed
            assert not (cache_dir / jintrac_file).exists()
            assert (cache_dir / solps_file).exists()

            # Check console output
            captured = capsys.readouterr()
            assert "iter_jintrac" in captured.out
            assert jintrac_file in captured.out

    def test_clear_cache_multiple_datasets_with_temp_directory(self, capsys):
        """Test _clear_cache with multiple datasets and temporary directory."""
        with tempfile.TemporaryDirectory() as temp_dir:
            cache_dir = Path(temp_dir) / "test_cache"
            cache_dir.mkdir(parents=True)

            # Create test files
            jintrac_file = method_files_map["iter_jintrac"][0]
            solps_file = method_files_map["iter_solps"][0]
            jorek_file = method_files_map["iter_jorek"][0]

            (cache_dir / jintrac_file).touch()
            (cache_dir / solps_file).touch()
            (cache_dir / jorek_file).touch()

            # Clear jintrac and solps datasets
            _clear_cache([iter_jintrac, iter_solps], cache_dir)

            # Verify jintrac and solps files are removed, jorek remains
            assert not (cache_dir / jintrac_file).exists()
            assert not (cache_dir / solps_file).exists()
            assert (cache_dir / jorek_file).exists()

    def test_clear_cache_nonexistent_cache_directory(self, capsys):
        """Test _clear_cache with non-existent cache directory."""
        with tempfile.TemporaryDirectory() as temp_dir:
            nonexistent_dir = Path(temp_dir) / "nonexistent"

            _clear_cache(None, nonexistent_dir)

            captured = capsys.readouterr()
            assert "doesn't exist" in captured.out
            assert "Nothing to clear" in captured.out

    def test_clear_cache_nonexistent_file(self, capsys):
        """Test _clear_cache when specific dataset files don't exist."""
        with tempfile.TemporaryDirectory() as temp_dir:
            cache_dir = Path(temp_dir) / "test_cache"
            cache_dir.mkdir(parents=True)

            # Don't create any files, just try to clear jintrac
            _clear_cache(iter_jintrac, cache_dir)

            captured = capsys.readouterr()
            assert "doesn't exist" in captured.out
            assert "Nothing to clear" in captured.out

    def test_clear_cache_invalid_dataset_type(self):
        """Test _clear_cache with invalid dataset type."""
        with tempfile.TemporaryDirectory() as temp_dir:
            cache_dir = Path(temp_dir) / "test_cache"
            cache_dir.mkdir(parents=True)

            with pytest.raises(TypeError, match="Expected a callable dataset method"):
                _clear_cache("invalid_dataset", cache_dir)

    def test_clear_cache_unknown_dataset_method(self):
        """Test _clear_cache with unknown dataset method."""

        def unknown_dataset():
            pass

        with tempfile.TemporaryDirectory() as temp_dir:
            cache_dir = Path(temp_dir) / "test_cache"
            cache_dir.mkdir(parents=True)

            with pytest.raises(ValueError, match="Dataset method unknown_dataset doesn't exist"):
                _clear_cache(unknown_dataset, cache_dir)

    def test_clear_cache_missing_platformdirs(self):
        """Test clear_cache behavior when platformdirs is not available."""
        with patch("cherab.imas.datasets._utils.platformdirs", None):
            with pytest.raises(ImportError, match="Missing optional dependency 'pooch'"):
                _clear_cache(None, None)

    @patch("cherab.imas.datasets._utils.platformdirs")
    def test_clear_cache_default_cache_dir(self, mock_platformdirs):
        """Test _clear_cache uses default cache directory when none provided."""
        mock_cache_dir = "/mock/cache/dir"
        mock_platformdirs.user_cache_dir.return_value = mock_cache_dir

        with patch("cherab.imas.datasets._utils.Path") as mock_path:
            mock_path_instance = MagicMock()
            mock_path.return_value = mock_path_instance
            mock_path_instance.exists.return_value = False

            _clear_cache(None, None)

            mock_platformdirs.user_cache_dir.assert_called_once_with("cherab/imas")
            mock_path.assert_called_once_with(mock_cache_dir)

    def test_clear_cache_custom_method_map(self, capsys):
        """Test _clear_cache with custom method map."""

        def custom_dataset():
            pass

        custom_method_map = {"custom_dataset": ["custom_file.nc"]}

        with tempfile.TemporaryDirectory() as temp_dir:
            cache_dir = Path(temp_dir) / "test_cache"
            cache_dir.mkdir(parents=True)

            custom_file = cache_dir / "custom_file.nc"
            custom_file.touch()

            _clear_cache(custom_dataset, cache_dir, custom_method_map)

            assert not custom_file.exists()
            captured = capsys.readouterr()
            assert "custom_dataset" in captured.out
            assert "custom_file.nc" in captured.out



# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Module for loading edge-profile-related data from IMAS IDS structures."""

from dataclasses import astuple, fields

import numpy as np
from numpy.typing import NDArray

from cherab.core.atomic import AtomicData
from imas.ids_primitive import IDSNumericArray
from imas.ids_structure import IDSStructArray, IDSStructure

from ..common import solve_coronal_equilibrium
from ..common.species import (
    ProfileData,
    SpeciesComposition,
    SpeciesData,
    SpeciesType,
    VelocityData,
    get_elements,
    get_ion,
    get_ion_state,
    get_neutral,
    get_neutral_state,
)

__all__ = [
    "load_edge_profiles",
    "load_edge_species",
]


def _get_profile(
    ids_struct: IDSStructure,
    name: str,
    grid_subset_index: int,
    name2: str | None = None,
) -> NDArray[np.float64] | None:
    struct_arr = getattr(ids_struct, name, None)
    if isinstance(struct_arr, IDSStructArray):
        for sub_struct in struct_arr:
            index = getattr(sub_struct, "grid_subset_index", None)
            if index is None:
                return None
            elif index == grid_subset_index:
                data = getattr(sub_struct, "values", None)
                if isinstance(data, IDSNumericArray):
                    return np.asarray(data) if len(data) else None
                elif data is None:
                    data = getattr(sub_struct, name2, None) if name2 is not None else None
                    if isinstance(data, IDSNumericArray):
                        return np.asarray(data) if len(data) else None
                    else:
                        return None
                else:
                    return None
        return None
    else:
        return None


def load_edge_profiles(
    species_struct: IDSStructure,
    species: SpeciesData,
    grid_subset_index: int = 5,
    backup_species_struct: IDSStructure | None = None,
) -> ProfileData:
    """Load edge profiles from a given species structure.

    Parameters
    ----------
    species_struct
        The ids structure containing the profiles for a single species.
    species
        The species data for the profiles to be loaded.
    grid_subset_index
        Identifier index of the grid subset, by default 5 (``"Cells"``).
    backup_species_struct
        The backup ids structure that is used if the profile is missing in species_struct,
        by default None.

    Returns
    -------
    `.ProfileData`
        Instance of the `.ProfileData` dataclass containing the loaded profiles.
    """
    profiles = ProfileData(species)
    velocities = VelocityData()

    for field in fields(profiles):
        match field.name:
            case "velocity":
                for field2 in fields(velocities):
                    data = _get_profile(
                        species_struct, field.name, grid_subset_index, name2=field2.name
                    )
                    if data is None and backup_species_struct is not None:
                        data = _get_profile(
                            backup_species_struct, field.name, grid_subset_index, name2=field2.name
                        )
                    setattr(velocities, field2.name, data)
                setattr(profiles, field.name, velocities)
            case "species":
                # species data is not stored in the profile structure, skip loading it here
                continue
            case _:
                data = _get_profile(species_struct, field.name, grid_subset_index)
                if data is None and backup_species_struct is not None:
                    data = _get_profile(backup_species_struct, field.name, grid_subset_index)
                setattr(profiles, field.name, data)

    return profiles


def load_edge_species(
    ggd_struct: IDSStructure,
    grid_subset_index: int = 5,
    split_ion_bundles: bool = True,
    atomic_data: AtomicData | None = None,
) -> SpeciesComposition:
    """Load edge plasma species and their profiles from a given GGD structure.

    Parameters
    ----------
    ggd_struct
        The ggd ids structure containing the profiles.
    grid_subset_index
        Identifier index of the grid subset, by default 5 (``"Cells"``).
    split_ion_bundles
        Whether to split ion bundles into individual ion states using `.solve_coronal_equilibrium`,
        by default True.
    atomic_data
        Optional atomic data to pass to `.solve_coronal_equilibrium` when splitting ion bundles.

    Returns
    -------
    `.SpeciesComposition`
        Instance of the `.SpeciesComposition` dataclass

    Raises
    ------
    RuntimeError
        If electron temperature or density profiles are missing, which are required for determining
        the species composition and solving coronal equilibrium when splitting ion bundles.
    RuntimeError
        If unable to determine the species due to missing element information, density profiles,
        or other necessary data.
    """
    composition = SpeciesComposition(
        electron=load_edge_profiles(
            ggd_struct.electrons, SpeciesData(z_min=-1, z_max=-1), grid_subset_index
        ),
    )

    if composition.electron.temperature is None:
        raise RuntimeError("Electron temperature profiles are required.")

    electron_density = composition.electron.density_thermal
    if electron_density is None:
        electron_density = composition.electron.density
    if electron_density is None:
        raise RuntimeError("Electron density or density_thermal profiles are required.")

    # Temporary sets
    ion_elements = []
    ion_uuids = set()
    ion_bundle_uuids = set()
    neutral_uuids = set()
    neutral_bundle_uuids = set()

    # ------------
    # === Ions ===
    # ------------
    for ion in ggd_struct.ion:
        ion: IDSStructure
        elements = get_elements(ion.element)
        if not len(elements):
            raise RuntimeError("Unable to determine the ion species, ion.element AOS is empty.")
        ion_elements.append(elements)

        # ---------------
        # === Bundles ===
        # ---------------
        if len(ion.state):
            shared_temperature = _get_profile(ion, "temperature", grid_subset_index)

            # Use the ion-level profiles as backup if there is only one state, otherwise no backup
            backup_ids = None if len(ion.state) > 1 else ion

            for i, state in enumerate(ion.state):
                species_data = get_ion_state(state, i, elements, grid_subset_index)
                species_tuple = astuple(species_data)
                if species_tuple in ion_bundle_uuids or species_tuple in ion_uuids:
                    print(f"Warning! Skipping duplicated ion: {species_data}")
                    continue

                profile_data = load_edge_profiles(
                    state, species_data, grid_subset_index, backup_ids
                )
                if backup_ids is None and profile_data.temperature is None:
                    profile_data.temperature = shared_temperature
                # === Case: ION ===
                if species_data.species_type == SpeciesType.ION:
                    composition.ion.append(profile_data)
                    ion_uuids.add(species_tuple)

                # === Case: ION_BUNDLE ===
                elif species_data.species_type == SpeciesType.ION_BUNDLE:
                    # Split ion bundles into individual ion states
                    if split_ion_bundles:
                        # Check if the necessary data is available before attempting to solve coronal equilibrium
                        elm = species_data.element or species_data.elements[0]
                        if elm is None:
                            raise RuntimeError(
                                f"Unable to determine the element for ion {species_data}, "
                                "cannot solve coronal equilibrium to split ion bundle."
                            )
                        density = profile_data.density_thermal
                        if density is None:
                            density = profile_data.density
                        if density is None:
                            raise RuntimeError(
                                f"Missing density profiles for ion {species_data}, "
                                "cannot solve coronal equilibrium to split ion bundle."
                            )
                        try:
                            densities_per_charge = solve_coronal_equilibrium(
                                elm,
                                density,
                                electron_density,
                                composition.electron.temperature,
                                atomic_data=atomic_data,
                                z_min=species_data.z_min,
                                z_max=species_data.z_max,
                            )
                        except Exception as e:
                            print(
                                f"Skipping ion bundle {species_data} "
                                f"due to error in solving coronal equilibrium: {e}"
                            )
                            composition.ion_bundle.append(profile_data)
                            ion_bundle_uuids.add(astuple(species_data))
                            continue

                        charge_states = np.arange(
                            species_data.z_min, species_data.z_max + 1, dtype=int
                        )

                        for i_charge, charge in enumerate(charge_states):
                            species = SpeciesData(
                                element=species_data.element,
                                z_min=charge,
                                z_max=charge,
                                species_type=SpeciesType.ION,
                            )
                            composition.ion.append(
                                ProfileData(
                                    species=species,
                                    density=densities_per_charge[i_charge],
                                    density_thermal=densities_per_charge[i_charge],
                                    temperature=profile_data.temperature,
                                    velocity=profile_data.velocity,
                                )
                            )
                            ion_uuids.add(astuple(species))

                    # Don't split ion bundles, just add the bundle as is
                    else:
                        composition.ion_bundle.append(profile_data)
                        ion_bundle_uuids.add(astuple(species_data))

                # === Case: Unexpected species type ===
                else:
                    print(f"Warning! Skipping ion with unexpected species {species_data}")

        # -------------------
        # === Non-bundled ===
        # -------------------
        else:
            species_data = get_ion(ion, elements)
            species_tuple = astuple(species_data)
            if species_tuple in ion_uuids:
                print(f"Warning! Skipping duplicated ion: {species_data}")
            else:
                profile_data = load_edge_profiles(ion, species_data, grid_subset_index)
                if species_data.species_type == SpeciesType.ION:
                    composition.ion.append(profile_data)
                    ion_uuids.add(species_tuple)
                else:
                    print(
                        f"Warning! Skipping non-bundled ion with unexpected species {species_data}"
                    )

    # ----------------------------
    # === Neutrals (molecules) ===
    # ----------------------------
    for neutral in ggd_struct.neutral:
        elements = get_elements(neutral.element)
        if not len(elements):
            elements = ion_elements[neutral.ion_index - 1]

        # ---------------
        # === Bundles ===
        # ---------------
        if len(neutral.state):
            shared_temperature = _get_profile(neutral, "temperature", grid_subset_index)

            # Use the neutral-level profiles as backup if there is only one state, otherwise no backup
            backup_ids = None if len(neutral.state) > 1 else neutral

            for state in neutral.state:
                species_data = get_neutral_state(state, elements)
                species_tuple = astuple(species_data)
                if species_tuple in neutral_bundle_uuids:
                    print(f"Warning! Skipping duplicated neutral: {species_data}")
                    continue
                neutral_bundle_uuids.add(species_tuple)

                profile_data = load_edge_profiles(
                    state, species_data, grid_subset_index, backup_ids
                )

                if backup_ids is None and profile_data.temperature is None:
                    profile_data.temperature = shared_temperature

                if species_data.species_type == SpeciesType.MOLECULAR_BUNDLE:
                    composition.molecular_bundle.append(profile_data)
                elif species_data.species_type == SpeciesType.MOLECULE:
                    composition.molecule.append(profile_data)
                elif species_data.species_type == SpeciesType.NEUTRAL_BUNDLE:
                    composition.neutral_bundle.append(profile_data)
                elif species_data.species_type == SpeciesType.NEUTRAL:
                    composition.neutral.append(profile_data)
                else:
                    print(f"Warning! Skipping neutral with unexpected species: {species_data}")

        # -------------------
        # === Non-bundled ===
        # -------------------
        else:
            species_data = get_neutral(neutral, elements)
            species_tuple = astuple(species_data)
            if species_tuple in neutral_uuids:
                print(f"Warning! Skipping duplicated neutral: {species_data}")
            else:
                profile_data = load_edge_profiles(neutral, species_data, grid_subset_index, None)
                if species_data.species_type == SpeciesType.MOLECULE:
                    composition.molecule.append(profile_data)
                    neutral_uuids.add(species_tuple)
                elif species_data.species_type == SpeciesType.NEUTRAL:
                    composition.neutral.append(profile_data)
                    neutral_uuids.add(species_tuple)
                else:
                    print(
                        f"Warning! Skipping non-bundled neutral with unexpected species: "
                        f"{species_data}"
                    )

    # Replace missing species temperature with average ion temperature
    t_ion = _get_profile(ggd_struct, "t_i_average", grid_subset_index)
    if t_ion is not None:
        species_types = {field.name for field in fields(composition)}
        species_types.remove("electron")
        for species_type in species_types:
            for profile in getattr(composition, species_type):
                if getattr(profile, "temperature", None) is None:
                    print(f"Warning! Using average ion temperature for the {profile.species}.")
                    profile.temperature = t_ion

    return composition



# cython: language_level=3

# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Module defining simple interpolators for data defined on a 3D unstructured grid."""
import numpy as np

from raysect.core.math.vector cimport new_vector3d
from raysect.core.math.point cimport new_point3d

cimport cython

__all__ = ["UnstructGridFunction3D", "UnstructGridVectorFunction3D"]


cdef class UnstructGridFunction3D(Function3D):
    """Simple interpolator for the data defined on the 3D unstructured grid.

    Find the cell containing the point (x, y, z) using the KDtree algorithm.
    Return the data value for this cell or the `fill_value` if the grid does not contain the point.

    Parameters
    ----------
    vertex_coords : (N,3) array_like
        3D array-like with the vertex coordinates of tetrahedra.
    tetrahedra : (M,4) array_like
        3D integer array-like with the vertex indices forming the tetrahedra.
    tetra_to_cell_map : (M,) array_like
        1D integer array-like with the indices of the grid cells (cube) containing the tetrahedra.
    grid_data : (L,) ndarray
        Array containing data in the grid cells.
    fill_value : float, optional
        Value returned outside the grid, by default 0.
    """

    def __init__(
        self,
        object vertex_coords not None,
        object tetrahedra not None,
        object tetra_to_cell_map not None,
        np.ndarray grid_data not None,
        double fill_value=0,
    ):

        vertex_coords = np.array(vertex_coords, dtype=np.float64)
        tetrahedra = np.array(tetrahedra, dtype=np.int32)
        tetra_to_cell_map = np.array(tetra_to_cell_map, dtype=np.int32)

        # build kdtree
        self._kdtree = MeshKDTree3D(vertex_coords, tetrahedra)

        self._tetra_to_cell_map = tetra_to_cell_map
        self._tetra_to_cell_map_mv = self._tetra_to_cell_map

        # NOTE: Attention!!! Do not copy grid_data!
        # Attribute self._grid_data must point to the original data array,
        # so as not to re-initialize the interpolator if the user changes data values.

        # populate internal attributes
        self._grid_data = grid_data
        self._fill_value = fill_value

        self._grid_data_mv = self._grid_data

    def __getstate__(self):
        return self._grid_data, self._fill_value, self._tetra_to_cell_map, self._kdtree

    def __setstate__(self, state):
        self._grid_data, self._fill_value, self._tetra_to_cell_map, self._kdtree = state
        self._tetra_to_cell_map_mv = self._tetra_to_cell_map
        self._grid_data_mv = self._grid_data

    def __reduce__(self):
        return self.__new__, (self.__class__, ), self.__getstate__()

    @classmethod
    def instance(cls, object instance not None, np.ndarray grid_data=None, object fill_value=None):
        """Create a new interpolator instance from an existing `UnstructGridFunction3D` or
        `UnstructGridVectorFunction3D` instance.

        The new interpolator instance will share the same internal acceleration
        data as the original interpolator. The grid_data of the new instance can
        be redefined.
        This method should be used if the user has multiple datasets
        that lie on the same mesh geometry. Using this methods avoids the
        repeated rebuilding of the mesh acceleration structures by sharing the
        geometry data between multiple interpolator objects.

        If created from the UnstructGridVectorFunction3D instance,
        the grid_data and the fill_value must not be None.

        Parameters
        ----------
        instance : UnstructGridFunction3D | UnstructGridVectorFunction3D
            The instance from which to create the new interpolator.
        grid_data : (L,) ndarray, optional
            Array containing data in the grid cells.
        fill_value : float, optional
            Value returned outside the grid, by default None.
            If None, inherited from the original instance.

        Returns
        -------
        UnstructGridFunction3D
            New interpolator instance.
        """

        cdef UnstructGridFunction3D m, inst
        cdef UnstructGridVectorFunction3D instvec

        m = UnstructGridFunction3D.__new__(UnstructGridFunction3D)

        if isinstance(instance, UnstructGridFunction3D):
            inst = instance
            # copy source data
            m._kdtree = inst._kdtree
            m._tetra_to_cell_map = inst._tetra_to_cell_map

            # replace grid data and fill value
            m._grid_data = inst._grid_data if grid_data is None else grid_data
            m._fill_value = inst._fill_value if fill_value is None else <double>fill_value
        elif isinstance(instance, UnstructGridVectorFunction3D):
            instvec = instance
            m._kdtree = instvec._kdtree
            m._tetra_to_cell_map = instvec._tetra_to_cell_map

            if grid_data is None:
                raise ValueError(
                    "Argument 'grid_data' must not be None "
                    "if the new instant UnstructGridFunction3D is created "
                    "from the UnstructGridVectorFunction3D instance."
                )
            if fill_value is None:
                raise ValueError(
                    "Argument 'fill_value' must not be None "
                    "if the new instant UnstructGridFunction3D is created "
                    "from the UnstructGridVectorFunction3D instance."
                )
            m._grid_data = grid_data
            m._fill_value = <double>fill_value
        else:
            raise TypeError(
                "Argument 'instance' must be either UnstructGridFunction3D "
                "or UnstructGridVectorFunction3D instance."
            )

        m._tetra_to_cell_map_mv = m._tetra_to_cell_map
        m._grid_data_mv = m._grid_data

        return m

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef double evaluate(self, double x, double y, double z) except? -1e999:

        cdef:
            np.int32_t tetrahedra_id, icell

        if self._kdtree.is_contained(new_point3d(x, y, z)):

            tetrahedra_id = self._kdtree.tetrahedra_id
            icell = self._tetra_to_cell_map_mv[tetrahedra_id]
            return self._grid_data_mv[icell]

        return self._fill_value

cdef class UnstructGridVectorFunction3D(VectorFunction3D):
    """Simple vector interpolator for the data defined on the 3D unstructured grid.

    Find the cell containing the point (x, y, z) using the KDtree algorithm.
    Return the 3D vector value for this cell or the `fill_vector` if the grid does not contain the
    point.

    Parameters
    ----------
    vertex_coords : (N,3) array_like
        3D array-like with the vertex coordinates of tetrahedra.
    tetrahedra : (M,4) array_like
        3D integer array-like with the vertex indices forming the tetrahedra.
    tetra_to_cell_map : (M,) array_like
        1D integer array-like with the indices of the grid cells (cube) containing the tetrahedra.
    grid_vectors : (3,L) ndarray
        Array containing 3D vectors in the grid cells.
    fill_vector : Vector3D, optional
        3D vector returned outside the grid, by default `Vector3D(0, 0, 0)`.
    """

    def __init__(
        self,
        object vertex_coords not None,
        object tetrahedra not None,
        object tetra_to_cell_map not None,
        np.ndarray grid_vectors not None,
        Vector3D fill_vector=Vector3D(0, 0, 0),
    ):

        vertex_coords = np.array(vertex_coords, dtype=np.float64)
        tetrahedra = np.array(tetrahedra, dtype=np.int32)
        tetra_to_cell_map = np.array(tetra_to_cell_map, dtype=np.int32)

        # build kdtree
        self._kdtree = MeshKDTree3D(vertex_coords, tetrahedra)

        self._tetra_to_cell_map = tetra_to_cell_map
        self._tetra_to_cell_map_mv = self._tetra_to_cell_map

        # NOTE: Attention!!! Do not copy grid_vectors!
        # Attribute self._grid_vectors must point to the original data array,
        # so as not to re-initialize the interpolator if the user changes data values.

        # populate internal attributes
        self._grid_vectors = grid_vectors
        self._fill_vector = fill_vector

        self._grid_vectors_mv = self._grid_vectors

    def __getstate__(self):
        return self._grid_vectors, self._fill_vector, self._tetra_to_cell_map, self._kdtree

    def __setstate__(self, state):
        self._grid_vectors, self._fill_vector, self._tetra_to_cell_map, self._kdtree = state
        self._grid_vectors_mv = self._grid_vectors
        self._tetra_to_cell_map_mv = self._tetra_to_cell_map

    def __reduce__(self):
        return self.__new__, (self.__class__, ), self.__getstate__()

    @classmethod
    def instance(
        cls,
        object instance not None,
        np.ndarray grid_vectors=None,
        Vector3D fill_vector=None,
    ):
        """Create a new interpolator instance from an existing `UnstructGridVectorFunction3D` or
        `UnstructGridFunction3D` instance.

        The new interpolator instance will share the same internal acceleration
        data as the original interpolator. The grid_vectors of the new instance can
        be redefined.
        This method should be used if the user has multiple datasets
        that lie on the same mesh geometry. Using this methods avoids the
        repeated rebuilding of the mesh acceleration structures by sharing the
        geometry data between multiple interpolator objects.

        If created from the UnstructGridFunction3D instance,
        the grid_vectors and the fill_vector must not be None.

        Parameters
        ----------
        instance : UnstructGridVectorFunction3D | UnstructGridFunction3D
            The instance from which to create the new interpolator.
        grid_vectors : (3,L) ndarray, optional
            Array containing vector grid data.
        fill_vector : Vector3D, optional
            3D vector returned outside the grid, by default None.
            If None, inherited from the original instance.

        Returns
        -------
        UnstructGridVectorFunction3D
            New interpolator instance.
        """

        cdef UnstructGridVectorFunction3D m, instvec
        cdef UnstructGridFunction3D inst

        m = UnstructGridVectorFunction3D.__new__(UnstructGridVectorFunction3D)

        if isinstance(instance, UnstructGridVectorFunction3D):
            instvec = instance
            # copy source data
            m._kdtree = instvec._kdtree
            m._tetra_to_cell_map = instvec._tetra_to_cell_map

            # replace grid vector and fill vector
            m._grid_vectors = instvec._grid_vectors if grid_vectors is None else grid_vectors
            m._fill_vector = instvec._fill_vector if fill_vector is None else fill_vector
        elif isinstance(instance, UnstructGridFunction3D):
            inst = instance
            m._kdtree = inst._kdtree
            m._tetra_to_cell_map = inst._tetra_to_cell_map

            if grid_vectors is None:
                raise ValueError(
                    "Argument 'grid_vectors' must not be None "
                    "if the new instant UnstructGridVectorFunction3D is created "
                    "from the UnstructGridFunction3D instance."
                )
            if fill_vector is None:
                raise ValueError(
                    "Argument 'fill_vector' must not be None "
                    "if the new instant UnstructGridVectorFunction3D is created "
                    "from the UnstructGridFunction3D instance."
                )
            m._grid_vectors = grid_vectors
            m._fill_vector = fill_vector
        else:
            raise TypeError(
                "Argument 'instance' must be either UnstructGridFunction3D "
                "or UnstructGridVectorFunction3D instance."
            )

        m._tetra_to_cell_map_mv = m._tetra_to_cell_map
        m._grid_vectors_mv = m._grid_vectors

        return m

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef Vector3D evaluate(self, double x, double y, double z):

        cdef:
            np.int32_t tetrahedra_id, icell
            double vx, vy, vz

        if self._kdtree.is_contained(new_point3d(x, y, z)):

            tetrahedra_id = self._kdtree.tetrahedra_id
            icell = self._tetra_to_cell_map_mv[tetrahedra_id]
            vx = self._grid_vectors_mv[0, icell]
            vy = self._grid_vectors_mv[1, icell]
            vz = self._grid_vectors_mv[2, icell]

            return new_vector3d(vx, vy, vz)

        return self._fill_vector



"""Module for solving some plasma models, such as the coronal equilibrium model."""

from __future__ import annotations

import numpy as np
from numpy.typing import ArrayLike, NDArray

from cherab.core.atomic import AtomicData, Element
from cherab.openadas import OpenADAS

__all__ = [
    "solve_coronal_equilibrium",
]


def solve_coronal_equilibrium(
    element: Element,
    density: ArrayLike,
    n_e: ArrayLike,
    t_e: ArrayLike,
    atomic_data: AtomicData | None = None,
    z_min: int | None = None,
    z_max: int | None = None,
) -> NDArray[np.float64]:
    r"""Solve the charge state distribution of an element in coronal equilibrium.

    The coronal equilibrium assumption is valid when the plasma is optically thin and the timescales
    for ionisation and recombination are much shorter than the timescales for transport processes.
    Under these conditions, the charge state distribution can be determined solely by the balance
    between ionisation and recombination processes.
    The charge state density of species :math:`Z`, denoted :math:`n_Z`, can be calculated from the
    following relation:

    .. math::

        n_{Z+1} = \frac{S_Z(n_\mathrm{e}, T_\mathrm{e})}{\alpha_{Z+1}(n_\mathrm{e}, T_\mathrm{e})}\cdot n_{Z},

    where :math:`S_Z(n_\mathrm{e}, T_\mathrm{e})` and :math:`\alpha_{Z+1}(n_\mathrm{e}, T_\mathrm{e})`
    are the ionisation and recombination rates coefficients for the charge state :math:`Z` as
    functions of the electron density :math:`n_\mathrm{e}` and temperature :math:`T_\mathrm{e}`,
    respectively.

    Using the total density of the element in the range of charge states considered i.e.,

    .. math::

        n_\mathrm{total}
            &= \sum_{Z=Z_\mathrm{min}}^{Z_\mathrm{max}} n_Z\\
            &= n_{Z_\mathrm{min}} + n_{Z_\mathrm{min}+1} + n_{Z_\mathrm{min}+2} + \dots\\
            &= n_{Z_\mathrm{min}}
               +
               \frac{S_{Z_\mathrm{min}}}
                    {\alpha_{Z_\mathrm{min}+1}}
               \cdot n_{Z_\mathrm{min}}
               +
               \frac{S_{Z_\mathrm{min}+1}}
                    {\alpha_{Z_\mathrm{min}+2}}
               \cdot
               \frac{S_{Z_\mathrm{min}}}
                    {\alpha_{Z_\mathrm{min}+1}}
               \cdot n_{Z_\mathrm{min}}
               +
               \dots\\
            &= n_{Z_\mathrm{min}}
               \cdot
               \left[
                 1
                 +
                 \sum_{i=0}^{Z_\mathrm{max} - Z_\mathrm{min} - 1}
                 \left(
                   \prod_{j=0}^{i}
                      \frac{S_{Z_\mathrm{min}+j}}
                           {\alpha_{Z_\mathrm{min}+j+1}}
                 \right)
               \right]

    allows us to express the first charge state density :math:`n_{Z_\mathrm{min}}` as

    .. math::

        n_{Z_\mathrm{min}}
            =
            n_\mathrm{total}
            \cdot
            \left[
                1 + \sum_{i=0}^{Z_\mathrm{max} - Z_\mathrm{min} - 1}
                    \left(
                        \prod_{j=0}^{i}\frac{S_{Z_\mathrm{min}+j}}{\alpha_{Z_\mathrm{min}+j+1}}
                    \right)
            \right]^{-1}.

    Substituting this back into the first equation :math:`n_{Z+1} = \frac{S_Z}{\alpha_{Z+1}} n_Z`
    allows us to compute the density of each charge state as follows:

    .. math::

        \begin{bmatrix} n_{Z_\mathrm{min}}\\ n_{Z_\mathrm{min}+1}\\ n_{Z_\mathrm{min}+2}\\ \vdots\\ n_{Z_\mathrm{max}} \end{bmatrix}
        =
        \frac{
          n_\mathrm{total}
        }{
          1 + \displaystyle\sum_{i=0}^{Z_\mathrm{max} - Z_\mathrm{min} - 1}
            \left(
                \prod_{j=0}^{i}\frac{S_{Z_\mathrm{min}+j}}{\alpha_{Z_\mathrm{min}+j+1}}
            \right)
        }
        \begin{bmatrix}
            1\\
            \frac{S_{Z_\mathrm{min}}}{\alpha_{Z_\mathrm{min}+1}}\\
            \frac{S_{Z_\mathrm{min}+1}}
                 {\alpha_{Z_\mathrm{min}+2}}
              \cdot
              \frac{S_{Z_\mathrm{min}}}{\alpha_{Z_\mathrm{min}+1}}\\
            \vdots\\
            \displaystyle\prod_{j=0}^{Z_\mathrm{max} - Z_\mathrm{min} - 1}
              \frac{S_{Z_\mathrm{min}+j}}{\alpha_{Z_\mathrm{min}+j+1}}
        \end{bmatrix}.

    Parameters
    ----------
    element
        Element for which to solve the charge state distribution.
    density
        Total density of the element, :math:`n_\mathrm{total}` (in m^-3).
        Should be the 1-D array of the same shape as `n_e` and `t_e`, or a scalar.
    n_e
        Electron density, :math:`n_\mathrm{e}` (in m^-3).
        Should be the 1-D array of the same shape as `density` and `t_e`, or a scalar.
    t_e
        Electron temperature, :math:`T_\mathrm{e}` (in eV).
        Should be the 1-D array of the same shape as `density` and `n_e`, or a scalar.
    atomic_data
        Atomic data provider to use for ionisation and recombination rates.
        If None, `~cherab.openadas.openadas.OpenADAS` will be used by default.
    z_min
        Minimum charge state to consider (inclusive), :math:`Z_\mathrm{min}`. If None, defaults to 0.
    z_max
        Maximum charge state to consider (inclusive), :math:`Z_\mathrm{max}`. If None, defaults to the atomic number of the element.

    Returns
    -------
    `numpy.ndarray`
        Density of the element in each charge state (in m^-3), with shape ``(z_max - z_min + 1, *density.shape)``.

    Raises
    ------
    ValueError
        If `z_min` or `z_max` are out of valid range, or if `density`, `n_e`, and `t_e` do not have the same shape.
    TypeError
        If `atomic_data` is not an instance of `~cherab.core.atomic.interface.AtomicData`.

    Examples
    --------
    >>> from cherab.core.atomic.elements import neon as element
    >>> density = 1.0  # if the density is 1, the output will be the relative abundance of each charge state
    >>> n_e = 1e20
    >>> t_e = 10
    >>> result = solve_coronal_equilibrium(element, density, n_e, t_e)
    >>> print([f"{x:.1%}" for x in result.ravel()])
    ['0.0%', '0.0%', '13.1%', '80.4%', '6.4%', '0.0%', '0.0%', '0.0%', '0.0%', '0.0%', '0.0%']
    """
    # Initialize variables if not provided
    z_min = z_min if z_min is not None else 0
    z_max = z_max if z_max is not None else element.atomic_number
    atomic_data = atomic_data if atomic_data is not None else OpenADAS(permit_extrapolation=True)

    # Validate z_min/z_max
    if z_min < 0:
        raise ValueError("z_min must be non-negative.")
    if z_max > element.atomic_number:
        raise ValueError("z_max cannot exceed the atomic number of the element.")
    if z_min > z_max:
        raise ValueError("z_min cannot be greater than z_max.")

    # Validate atomic_data
    if not isinstance(atomic_data, AtomicData):
        raise TypeError("atomic_data must be an instance of AtomicData.")

    # Validate array-like inputs
    density = np.atleast_1d(density)
    n_e = np.atleast_1d(n_e)
    t_e = np.atleast_1d(t_e)

    if density.shape != n_e.shape or density.shape != t_e.shape:
        raise ValueError("density, n_e, and t_e must have the same shape.")

    # Define charge states and compute ionisation/recombination ratios
    charges = np.arange(z_min, z_max + 1, dtype=int)
    ratios = np.zeros((len(charges), len(density)))

    for i_charge in range(len(charges) - 1):
        rate_ion = atomic_data.ionisation_rate(element, charges[i_charge])
        rate_rec = atomic_data.recombination_rate(element, charges[i_charge] + 1)

        for i_density in range(density.size):
            ratios[i_charge, i_density] = rate_ion(n_e[i_density], t_e[i_density]) / rate_rec(
                n_e[i_density], t_e[i_density]
            )

    # Compute fractional abundances of each charge state
    fractions = np.vstack(
        [
            np.ones((1, ratios.shape[1]), dtype=float),
            np.cumprod(ratios[:-1, :], axis=0, dtype=float),
        ]
    )

    # Normalize
    fractions /= np.sum(fractions, axis=0)

    # Return the charge state densities by multiplying the fractions with the total density
    return fractions * density



"""Utility functions for cherab.imas.datasets module."""

import shutil
from collections.abc import Callable, Collection
from pathlib import Path

from ._registry import method_files_map

try:
    import platformdirs
except ImportError:
    platformdirs = None


def _clear_cache(
    datasets: Callable[..., str] | Collection[Callable[..., str]] | None,
    cache_dir: str | Path | None = None,
    method_map: dict[str, list[str]] | None = None,
) -> None:
    if method_map is None:
        # Use Datasets method map
        method_map = method_files_map
    if cache_dir is None:
        # Use default cache_dir path
        if platformdirs is None:
            # platformdirs is pooch dependency
            raise ImportError(
                "Missing optional dependency 'pooch' required for cherab.imas.datasets module. "
                + "Please use pip or conda to install 'pooch'."
            )
        cache_dir = platformdirs.user_cache_dir("cherab/imas")

    # Ensure cache_dir is a Path object
    cache_dir = Path(cache_dir)

    if not cache_dir.exists():
        print(f"Cache Directory {cache_dir} doesn't exist. Nothing to clear.")
        return

    if datasets is None:
        print(f"Cleaning the cache directory {cache_dir}!")
        shutil.rmtree(cache_dir)
    else:
        if not isinstance(datasets, Collection):
            # single dataset method passed should be converted to list
            datasets = [
                datasets,
            ]
        for dataset in datasets:
            if not callable(dataset):
                raise TypeError(
                    "Expected a callable dataset method or a list/tuple of the same, "
                    + f"but got {type(dataset)}"
                )
            dataset_name = dataset.__name__  # Name of the dataset method
            if dataset_name not in method_map:
                raise ValueError(
                    f"Dataset method {dataset_name} doesn't exist."
                    + "Please check if the passed dataset is a subset of the following dataset "
                    + f"methods: {list(method_map.keys())}"
                )

            data_files = method_map[dataset_name]
            data_filepaths = [cache_dir / file for file in data_files]
            for data_filepath in data_filepaths:
                if data_filepath.exists():
                    print(f"Cleaning the file {data_filepath.name} for dataset {dataset_name}")
                    data_filepath.unlink()
                else:
                    print(f"Path {data_filepath} doesn't exist. Nothing to clear.")


def clear_cache(
    datasets: Callable[..., str] | Collection[Callable[..., str]] | None = None,
) -> None:
    """Clean the cherab-imas datasets cache directory.

    If a `cherab.imas.datasets` method or a list/tuple of the same is provided, then `clear_cache`
    removes all the data files associated to the passed dataset method callable(s).

    By default, it removes all the cached data files.

    Parameters
    ----------
    datasets
        `cherab.imas.datasets` method or a list/tuple of the same whose cached data files are to be
        removed. If `None`, all the cached data files are removed.

    Examples
    --------
    >>> from cherab.imas import datasets
    >>> data_path = datasets.iter_jintrac()
    >>> datasets.clear_cache([datasets.iter_jintrac])
    Cleaning the file iter_scenario_53298_seq1_DD4.nc for dataset iter_jintrac

    If no argument is passed, all cached dataset files are removed.

    >>> datasets.clear_cache()
    Cleaning the cache directory ~/Library/Caches/cherab/imas!
    """
    _clear_cache(datasets)



# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Module to offer the plasma object, blending core with edge profiles."""

from collections.abc import Callable
from dataclasses import fields

import numpy as np
from raysect.core.math import translate
from raysect.core.math.function.float import Blend2D, Blend3D, Function2D, Function3D
from raysect.core.math.function.vector3d import Blend2D as VectorBlend2D
from raysect.core.math.function.vector3d import Blend3D as VectorBlend3D
from raysect.core.math.function.vector3d import Function2D as VectorFunction2D
from raysect.core.math.function.vector3d import Function3D as VectorFunction3D
from raysect.core.scenegraph._nodebase import _NodeBase
from raysect.primitive import Cylinder, Subtract
from scipy.constants import atomic_mass, electron_mass

from cherab.core import AtomicData, Maxwellian, Plasma, Species
from cherab.core.math import AxisymmetricMapper, VectorAxisymmetricMapper
from cherab.tools.equilibrium import EFITEquilibrium
from imas import DBEntry
from imas.ids_structure import IDSStructure

from ..ids.common import get_ids_time_slice
from ..ids.common.ggd import load_grid
from ..ids.core_profiles import load_core_grid, load_core_species
from ..ids.edge_profiles import load_edge_species
from .core import get_core_interpolators, get_psi_norm, load_core_plasma
from .edge import get_edge_interpolators, load_edge_plasma
from .equilibrium import load_equilibrium, load_magnetic_field
from .utility import (
    ZERO_VELOCITY,
    ProfileInterpolator,
    get_subset_name_index,
    warn_unsupported_species,
)

__all__ = ["load_plasma"]


def load_plasma(
    *args,
    time: float = 0,
    occurrence_core: int = 0,
    edge_args: tuple | None = None,
    edge_kwargs: dict | None = None,
    time_edge: float | None = None,
    occurrence_edge: int = 0,
    grid_ggd: IDSStructure | None = None,
    grid_subset_id: int | str = 5,
    equilibrium: EFITEquilibrium | None = None,
    b_field: VectorFunction2D | None = None,
    psi_interpolator: Callable[[float], float] | None = None,
    mask: Function2D | Function3D | None = None,
    time_threshold: float = np.inf,
    split_ion_bundles: bool = True,
    atomic_data: AtomicData | None = None,
    parent: _NodeBase | None = None,
    **kwargs,
) -> Plasma:
    """Load core and edge profiles and create a `~cherab.core.plasma.node.Plasma` object.

    If the ``edge_profiles`` IDS is empty, returns only the core plasma.
    If the ``core_profiles`` IDS is empty, returns only the edge plasma.

    To load the edge plasma from a different IMAS entry, use `edge_args` and `edge_kwargs`
    to pass different arguments to the `~imas.db_entry.DBEntry` constructor.

    The distribution of each species is defined with `~cherab.core.distribution.Maxwellian` using
    its density and temperature profiles, which are mapped to 3D using the provided `equilibrium`.

    The ion bundle species are split into their constituent charge states
    using `.solve_coronal_equilibrium` when `.split_ion_bundles` is True and the necessary atomic
    data is available. Otherwise, ion bundles are ignored with a warning.

    Parameters
    ----------
    *args
        Arguments passed to the `~imas.db_entry.DBEntry` constructor.
    time
        Time for the core plasma, by default 0.
    occurrence_core
        Occurrence index of the ``core_profiles`` IDS, by default 0.
    edge_args
        Arguments passed to the `~imas.db_entry.DBEntry` constructor for the edge plasma
        if different from the core plasma. By default None: uses the same as `*args`.
    edge_kwargs
        Keyword arguments passed to the `~imas.db_entry.DBEntry` constructor for the edge plasma
        if different from the core plasma. By default None: uses the same as `**kwargs`.
    time_edge
        Time for the edge plasma. If None, uses `~cherab.imas.plasma.load_plasma.time`.
        By default None.
    occurrence_edge
        Occurrence index of the ``edge_profiles`` IDS, by default 0.
    grid_ggd
        Alternative ``grid_ggd`` structure describing the grid. By default None.
    grid_subset_id
        Identifier of the grid subset (index or name). By default 5 (``"Cells"``).
    equilibrium
        Alternative `~cherab.tools.equilibrium.efit.EFITEquilibrium` used to map core
        profiles. By default None: the equilibrium is read from the same IMAS query as the
        core profiles. Ignored if the core plasma is not available.
    b_field
        Alternative 2D interpolator of the magnetic field vector (Br, Bphi, Bz).
        By default None: the magnetic field is loaded from the ``equilibrium`` IDS.
    psi_interpolator
        Alternative ``psi_norm(rho_tor_norm)`` interpolator.
        Used only if ``psi`` is missing in the core grid. By default None.
        Obtained from the ``equilibrium`` IDS.
    mask
        Mask function used for blending: ``(1 - mask) * f_edge + mask * f_core``.
        By default, uses `~cherab.tools.equilibrium.efit.EFITEquilibrium`'s `inside_lcfs`.
    time_threshold
        Maximum allowed difference between the requested time and the nearest
        available time, by default `numpy.inf`.
    split_ion_bundles
        Whether to split ion bundles into their constituent charge states using
        `.solve_coronal_equilibrium`, by default True.
    atomic_data
        Atomic data provider class for this plasma, by default None.
        If None, some species (e.g. ion_bundle) may not be properly loaded.
    parent
        Parent node in the Raysect scene graph, by default None.
        Typically a `~raysect.optical.scenegraph.world.World` instance.
    **kwargs
        Keyword arguments passed to the `~imas.db_entry.DBEntry` constructor.

    Returns
    -------
    `~cherab.core.plasma.node.Plasma`
        Plasma object with core and/or edge profiles.

    Raises
    ------
    RuntimeError
        If neither core nor edge profiles are available in the IMAS entry.
    ValueError
        If the provided equilibrium is not an instance of `~cherab.tools.equilibrium.efit.EFITEquilibrium`.
    RuntimeError
        If there are issues with the core or edge profiles (e.g. missing electron profile,
        missing density or temperature profiles, empty grids, etc.) that prevent the plasma from
        being created.
    """
    # -----------------------------------------
    # === Load core/edge profiles from IMAS ===
    # -----------------------------------------
    # The core and edge profiles are loaded separately to allow for the case where one of them is missing.
    # If the edge profiles are missing, the core plasma is still created and returned, and vice versa.

    edge_args = edge_args or args
    edge_kwargs = edge_kwargs or kwargs
    if time_edge is None:
        time_edge = time

    # === Core profiles IDS ===
    try:
        with DBEntry(*args, **kwargs) as entry_core:
            core_profiles_ids = get_ids_time_slice(
                entry_core,
                "core_profiles",
                time=time,
                occurrence=occurrence_core,
                time_threshold=time_threshold,
            )
    except RuntimeError:
        return load_edge_plasma(
            *edge_args,
            time=time_edge,
            occurrence=occurrence_edge,
            grid_ggd=grid_ggd,
            grid_subset_id=grid_subset_id,
            b_field=b_field,
            time_threshold=time_threshold,
            split_ion_bundles=split_ion_bundles,
            atomic_data=atomic_data,
            parent=parent,
            **edge_kwargs,
        )

    # === Edge profiles IDS ===
    try:
        with DBEntry(*edge_args, **edge_kwargs) as entry_edge:
            edge_profiles_ids = get_ids_time_slice(
                entry_edge,
                "edge_profiles",
                time=time_edge,
                occurrence=occurrence_edge,
                time_threshold=time_threshold,
            )
    except RuntimeError:
        return load_core_plasma(
            *args,
            time=time,
            occurrence=occurrence_core,
            equilibrium=equilibrium,
            b_field=b_field,
            psi_interpolator=psi_interpolator,
            time_threshold=time_threshold,
            parent=parent,
            split_ion_bundles=split_ion_bundles,
            atomic_data=atomic_data,
            **kwargs,
        )

    if not len(core_profiles_ids.profiles_1d):
        raise RuntimeError("The profiles_1d AOS in core_profiles IDS is empty.")

    if not len(edge_profiles_ids.grid_ggd) and grid_ggd is None:
        raise RuntimeError(
            "The 'grid_ggd' AOS of the edge_profiles IDS is empty "
            + "and an alternative grid_ggd structure is not provided."
        )

    if not len(edge_profiles_ids.ggd):
        raise RuntimeError("The 'ggd' AOS of the edge_profiles IDS is empty.")

    # Load equilibrium and magnetic field data
    if equilibrium is None:
        equilibrium, psi_interp = load_equilibrium(
            *args, time=time, with_psi_interpolator=True, **kwargs
        )
        psi_interpolator = psi_interpolator or psi_interp

    if not isinstance(equilibrium, EFITEquilibrium):
        raise ValueError("Argument equilibrium must be a EFITEquilibrium instance.")

    if b_field is None:
        try:
            b_field = load_magnetic_field(*args, time=time, **kwargs)
        except RuntimeError:
            b_field = equilibrium.b_field

    # If no mask is provided, use the inside_lcfs function of the equilibrium as the default mask.
    mask = mask or equilibrium.inside_lcfs

    # === Core grid, composition and psi_norm ===
    core_grid = load_core_grid(core_profiles_ids.profiles_1d[0].grid)

    composition_core = load_core_species(
        core_profiles_ids.profiles_1d[0],
        split_ion_bundles=split_ion_bundles,
        atomic_data=atomic_data,
    )

    psi_norm = get_psi_norm(
        core_grid.psi,
        equilibrium.psi_axis,
        equilibrium.psi_lcfs,
        core_grid.rho_tor_norm,
        psi_interpolator,
    )

    # === Edge grid and composition ===
    grid_ggd = grid_ggd or edge_profiles_ids.grid_ggd[0]
    grid, subsets, subset_id = load_grid(grid_ggd, with_subsets=True)

    grid_subset_name, grid_subset_index = get_subset_name_index(subset_id, grid_subset_id)

    if np.all(subsets[grid_subset_name] != np.arange(grid.num_cell, dtype=int)):
        # To reduce memory usage, create the sub-grid only if needed.
        grid = grid.subset(subsets[grid_subset_name], name=grid_subset_name)

    composition_edge = load_edge_species(
        edge_profiles_ids.ggd[0],
        grid_subset_index=grid_subset_index,
        split_ion_bundles=split_ion_bundles,
        atomic_data=atomic_data,
    )

    # ----------------------------
    # === Create Plasma object ===
    # ----------------------------
    time_core = core_profiles_ids.time[0]
    time_edge = edge_profiles_ids.time[0]
    name = (
        f"IMAS core + edge plasma: core/edge time {time_core}/{time_edge}, "
        f"uri {entry_core.uri} / {entry_edge.uri}."
    )
    plasma = Plasma(parent=parent, name=name)

    # Create plasma geometry
    radius_outer = grid.mesh_extent["rmax"]
    radius_inner = grid.mesh_extent["rmin"]
    height = grid.mesh_extent["zmax"] - grid.mesh_extent["zmin"]
    zmin = grid.mesh_extent["zmin"]
    plasma.geometry = Subtract(Cylinder(radius_outer, height), Cylinder(radius_inner, height))
    plasma.geometry_transform = translate(0, 0, zmin)

    # Add magnetic field
    plasma.b_field = VectorAxisymmetricMapper(b_field)

    # Add atomic data
    if atomic_data is not None:
        plasma.atomic_data = atomic_data

    # ------------------------------------
    # === Define Electron Distribution ===
    # ------------------------------------
    for composition, ids_name in [
        (composition_core, "core_profiles"),
        (composition_edge, "edge_profiles"),
    ]:
        if composition.electron.density_thermal is not None:
            composition.electron.density = composition.electron.density_thermal

        if composition.electron is None:
            raise RuntimeError(f"Electron profile is missing in the {ids_name} IDS.")
        if composition.electron.density is None:
            raise RuntimeError(f"Electron density profile is missing in the {ids_name} IDS.")
        if composition.electron.temperature is None:
            raise RuntimeError(f"Electron temperature profile is missing in the {ids_name} IDS.")

    interp_edge = get_edge_interpolators(grid, composition_edge.electron, b_field, return3d=True)
    interp_core = get_core_interpolators(
        psi_norm,
        composition_core.electron,
        equilibrium,
        return3d=True,
    )

    interp = blend_core_edge_interpolators(interp_core, interp_edge, mask, return3d=True)

    plasma.electron_distribution = Maxwellian(
        interp.density,
        interp.temperature,
        interp.velocity or ZERO_VELOCITY,
        electron_mass,
    )

    # -----------------------------------------------
    # === Define Species Distribution/Composition ===
    # -----------------------------------------------

    # === Ion/Neutral Species ===

    # Validate and prepare the core and edge species profiles
    species_core = {}
    species_edge = {}
    for composition, space in [
        (composition_core, "core"),
        (composition_edge, "edge"),
    ]:
        for profile in composition.ion + composition.neutral:
            if profile.species is None:
                print(
                    f"Warning! Skipping {space} species with missing element or charge: {profile}"
                )
                continue
            if profile.species.element is None:
                print(f"Warning! Skipping {space} species with missing element: {profile}")
                continue
            if profile.density_thermal is not None:
                profile.density = profile.density_thermal
            if profile.density is None:
                print(f"Warning! Skipping {space} {profile.species}: density profile is missing.")
                continue
            if profile.temperature is None:
                print(
                    f"Warning! Skipping {space} {profile.species}: temperature profile is missing."
                )
                continue

            element = profile.species.element
            charge = profile.species.z_min

            # Store the profile temporarily
            sp_key = (element, charge)
            if space == "core":
                species_core[sp_key] = profile
            else:
                species_edge[sp_key] = profile

    # Blend core and edge species profiles together
    species = {}

    for core_key, core_profile in species_core.items():
        if core_key in species_edge:
            edge_profiles = species_edge[core_key]

            core_interp = get_core_interpolators(
                psi_norm, core_profile, equilibrium, return3d=False
            )
            edge_interp = get_edge_interpolators(grid, edge_profiles, b_field, return3d=False)

            species[core_key] = blend_core_edge_interpolators(
                core_interp, edge_interp, mask, return3d=True
            )
        else:
            species[core_key] = get_core_interpolators(
                psi_norm, core_profile, equilibrium, return3d=True
            )

    for edge_key, edge_profiles in species_edge.items():
        if edge_key not in species_core:
            species[edge_key] = get_edge_interpolators(grid, edge_profiles, b_field, return3d=True)

    # Add the blended species to the plasma composition
    for (element, charge), interp in species.items():
        try:
            species = plasma.composition.get(element, int(charge))
            print(f"Warning! Skipping {species}: already defined")
            continue
        except ValueError:
            pass

        distribution = Maxwellian(
            interp.density,
            interp.temperature,
            interp.velocity or ZERO_VELOCITY,
            element.atomic_weight * atomic_mass,
        )

        plasma.composition.add(Species(element, int(charge), distribution))

    # === Ion Bundle Species ===
    # Ion bundles are split into their constituent charge states at the composition level.
    warn_unsupported_species(composition_core, "ion_bundle")
    warn_unsupported_species(composition_edge, "ion_bundle")

    # === Molecular Species ===
    # TODO: properly support molecular species.
    # For now, just issue a warning if any molecule or molecular_bundle species are present in the composition.
    warn_unsupported_species(composition_core, "molecule")
    warn_unsupported_species(composition_edge, "molecule")
    warn_unsupported_species(composition_core, "molecular_bundle")
    warn_unsupported_species(composition_edge, "molecular_bundle")

    return plasma


def blend_core_edge_interpolators(
    core_interpolators: ProfileInterpolator,
    edge_interpolators: ProfileInterpolator,
    mask: Function2D | Function3D,
    return3d: bool = False,
) -> ProfileInterpolator:
    """Blend together interpolators for the core and edge using the modulating mask function.

    Parameters
    ----------
    core_interpolators
        Instance of `.ProfileInterpolator` with 2D or 3D core profiles interpolators.
    edge_interpolators
        Instance of `.ProfileInterpolator` with 2D or 3D edge profiles interpolators.
    mask
        Mask function used for blending: ``(1 - mask) * f_edge + mask * f_core``.
    return3d
        If True, return the 3D functions for 2D interpolators assuming rotational symmetry,
        by default False.

    Returns
    -------
    ProfileInterpolator
        Instance of `.ProfileInterpolator` with blended interpolators.
    """
    interpolators = ProfileInterpolator()

    for field in fields(core_interpolators):
        core_func = getattr(core_interpolators, field.name, None)
        edge_func = getattr(edge_interpolators, field.name, None)
        setattr(
            interpolators,
            field.name,
            _blend_core_edge_functions(core_func, edge_func, mask, return3d),
        )

    return interpolators


def _blend_core_edge_functions(
    core_func: Function2D | Function3D | VectorFunction2D | VectorFunction3D | None,
    edge_func: Function2D | Function3D | VectorFunction2D | VectorFunction3D | None,
    mask: Function2D | Function3D,
    return3d: bool,
) -> Function2D | Function3D | VectorFunction2D | VectorFunction3D | None:
    """Blend together the core and edge interpolating functions using the modulating mask function.

    Parameters
    ----------
    core_func
        A 2D or 3D core interpolator.
    edge_func
        A 2D or 3D edge interpolator.
    mask
        The 2D or 3D mask function used for blending: ``(1 - mask) * f_edge + mask * f_core``.
    return3d
        If True, return the 3D functions for 2D interpolators assuming
        rotational symmetry, by default False.

    Returns
    -------
    Function2D | Function3D | VectorFunction2D | VectorFunction3D | None
        The blended function, or None if both input functions are None.

    Raises
    ------
    TypeError
        If the input functions are not 2D or 3D scalar/vector functions.
    RuntimeError
        If the core and edge functions have incompatible dimensions or types for blending.
    """
    if core_func is None and edge_func is None:
        return None

    # === Validation ===
    if core_func is not None and not isinstance(
        core_func, Function2D | Function3D | VectorFunction2D | VectorFunction3D
    ):
        raise TypeError("The core_func must be a 2D or 3D function.")

    if edge_func is not None and not isinstance(
        edge_func, Function2D | Function3D | VectorFunction2D | VectorFunction3D
    ):
        raise TypeError("The edge_func must be a 2D or 3D function.")

    if not isinstance(mask, Function2D | Function3D):
        raise TypeError("The mask must be a 2D or 3D function.")

    # === Only one of the two functions is available ===
    if core_func is None:
        if isinstance(edge_func, Function2D) and return3d:
            return AxisymmetricMapper(edge_func)
        elif isinstance(edge_func, VectorFunction2D) and return3d:
            return VectorAxisymmetricMapper(edge_func)
        else:
            return edge_func

    if edge_func is None:
        if isinstance(core_func, Function2D) and return3d:
            return AxisymmetricMapper(core_func)
        elif isinstance(core_func, VectorFunction2D) and return3d:
            return VectorAxisymmetricMapper(core_func)
        else:
            return core_func

    # === Both functions are available: blend them together ===
    if (
        isinstance(core_func, Function2D)
        and isinstance(edge_func, Function2D)
        and isinstance(mask, Function2D)
    ):
        blended_func = Blend2D(edge_func, core_func, mask)
        return AxisymmetricMapper(blended_func) if return3d else blended_func

    if (
        isinstance(core_func, VectorFunction2D)
        and isinstance(edge_func, VectorFunction2D)
        and isinstance(mask, Function2D)
    ):
        blended_func = VectorBlend2D(edge_func, core_func, mask)
        return VectorAxisymmetricMapper(blended_func) if return3d else blended_func

    # unable to return 2D, convert to 3D

    if isinstance(core_func, Function2D):
        core_func = AxisymmetricMapper(core_func)

    if isinstance(core_func, VectorFunction2D):
        core_func = VectorAxisymmetricMapper(core_func)

    if isinstance(edge_func, Function2D):
        edge_func = AxisymmetricMapper(edge_func)

    if isinstance(edge_func, VectorFunction2D):
        edge_func = VectorAxisymmetricMapper(edge_func)

    if isinstance(mask, Function2D):
        mask = AxisymmetricMapper(mask)

    if isinstance(core_func, Function3D) and isinstance(edge_func, Function3D):
        return Blend3D(edge_func, core_func, mask)

    if isinstance(core_func, VectorFunction3D) and isinstance(edge_func, VectorFunction3D):
        return VectorBlend3D(edge_func, core_func, mask)

    raise RuntimeError("Cannot blend scalar and vector functions.")



#!/usr/bin/env -S pixi run python
"""Core plasma profile plotting demo for ITER scenarios.

This demo creates a Plasma object from the core profiles and equilibrium. Then samples and plots the
quantities.
"""

from pathlib import Path

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.colors import SymLogNorm
from matplotlib.figure import Figure

from cherab.core.math import sample3d_grid, samplevector2d
from cherab.imas.datasets import iter_jintrac
from cherab.imas.plasma import load_core_plasma, load_equilibrium
from cherab.tools.equilibrium import plot_equilibrium

plt.ion()


def plot_quantity(quantity, extent, title="", logscale=False, symmetric=False) -> Figure:
    """Make a 2D plot of quantity, with a title, optionally on a log scale.

    Parameters
    ----------
    quantity
        2D array of the quantity to plot.
    extent
        The extent of the plot in the form [xmin, xmax, ymin, ymax].
    title
        The title of the plot.
    logscale
        Whether to use a logarithmic scale for the color map.
    symmetric
        Whether to use a symmetric color map around zero.

    Returns
    -------
    The matplotlib Figure object containing the plot.
    """
    fig = plt.figure(figsize=(4.0, 6.0), layout="constrained")
    ax = fig.add_subplot(111)
    if logscale:
        # Plot lowest values (mainly 0's) on linear map, as log(0) = -inf.
        linthresh = np.percentile(np.unique(quantity), 1)
        norm = SymLogNorm(linthresh=float(max(linthresh, 1.0e-10 * quantity.max())), base=10)
    else:
        norm = None
    # Sampled data is indexed as quantity(x, y), but matplotlib's imshow
    # expects quantity(y, x).
    if symmetric and not logscale:
        vmax = np.abs(quantity).max()
        image = ax.imshow(
            quantity.T, extent=extent, origin="lower", vmin=-vmax, vmax=vmax, cmap="coolwarm"
        )
    else:
        image = ax.imshow(quantity.T, extent=extent, origin="lower", norm=norm, cmap="gnuplot")
    fig.colorbar(image, aspect=50)
    ax.set_xlim(extent[0], extent[1])
    ax.set_ylim(extent[2], extent[3])
    ax.set_xlabel("R [m]")
    ax.set_ylabel("Z [m]")
    ax.set_title(title)

    return fig


demos_path = Path(__file__).parent
plots_path = demos_path / "plots"
plots_path.mkdir(exist_ok=True)

# Load and plot equilibrium
path = iter_jintrac()
equilibrium, psi_interpolator = load_equilibrium(path, "r", with_psi_interpolator=True)
plot_equilibrium(equilibrium)
plt.gcf().savefig(plots_path / "equilibrium.png", dpi=200)

# sampling range
xl, xu = equilibrium.r_range
zl, zu = equilibrium.z_range
nz = 600
nx = 300
extent = [xl, xu, zl, zu]

# Sample and plot magnetic field
xsamp, zsamp, b = samplevector2d(equilibrium.b_field, (xl, xu, nx), (zl, zu, nz))
b_length = np.sqrt((b * b).sum(2))
fig = plot_quantity(b[:, :, 0], extent, title="Brad [T]", symmetric=True)
fig.savefig(plots_path / "brad.png", dpi=200)
fig = plot_quantity(b[:, :, 1], extent, title="Btor [T]")
fig.savefig(plots_path / "btor.png", dpi=200)
fig = plot_quantity(b[:, :, 2], extent, title="Bz [T]", symmetric=True)
fig.savefig(plots_path / "bz.png", dpi=200)

# Load core plasma
plasma = load_core_plasma(path, "r", equilibrium=equilibrium, psi_interpolator=psi_interpolator)

# Sample and plot electron profiles
ne_plasma = sample3d_grid(plasma.electron_distribution.density, xsamp, [0], zsamp)
ne_plasma = ne_plasma.squeeze()
fig = plot_quantity(ne_plasma, extent, title="ne [m-3]")
fig.savefig(plots_path / "core_ne.png", dpi=200)

te_plasma = sample3d_grid(
    plasma.electron_distribution.effective_temperature, xsamp, [0], zsamp
).squeeze()
fig = plot_quantity(te_plasma, extent, title="Te [eV]")
fig.savefig(plots_path / "core_te.png", dpi=200)

# Sample and plot ion and neutral profiles
for species in plasma.composition:
    density = sample3d_grid(species.distribution.density, xsamp, [0], zsamp).squeeze()
    if np.any(density):
        fig = plot_quantity(
            density,
            extent,
            title=f"{species.element.symbol}{species.charge} density [m-3]",
        )
        fig.savefig(
            plots_path / f"core_{species.element.symbol}{species.charge}_density.png",
            dpi=200,
        )

    if species.element.atomic_number == 1:
        temperature = sample3d_grid(
            species.distribution.effective_temperature, xsamp, [0], zsamp
        ).squeeze()
        if np.any(temperature):
            fig = plot_quantity(
                temperature,
                extent,
                title=f"{species.element.symbol}{species.charge} temperature [eV]",
            )
            fig.savefig(
                plots_path / f"core_{species.element.symbol}{species.charge}_temperature.png",
                dpi=200,
            )



#!/usr/bin/env -S pixi run python
"""Edge plasma profile plotting demo for ITER-like scenarios.

This demo creates a Plasma object from the edge profiles. Then samples and plots the quantities.
"""

from pathlib import Path

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.colors import SymLogNorm
from matplotlib.figure import Figure

from cherab.core.math import sample3d, sample3d_grid, samplevector2d, samplevector3d_grid
from cherab.imas.datasets import iter_jintrac
from cherab.imas.plasma import load_edge_plasma, load_equilibrium, load_magnetic_field
from cherab.tools.equilibrium import plot_equilibrium

plt.ion()


def plot_quantity(quantity, extent, title="", logscale=False, symmetric=False) -> Figure:
    """Make a 2D plot of quantity, with a title, optionally on a log scale.

    Parameters
    ----------
    quantity
        2D array of the quantity to plot.
    extent
        The extent of the plot in the form [xmin, xmax, ymin, ymax].
    title
        The title of the plot.
    logscale
        Whether to use a logarithmic scale for the color map.
    symmetric
        Whether to use a symmetric color map around zero.

    Returns
    -------
    The matplotlib Figure object containing the plot.
    """
    fig = plt.figure(figsize=(4.0, 6.0), layout="constrained")
    ax = fig.add_subplot(111)
    if logscale:
        # Plot lowest values (mainly 0's) on linear map, as log(0) = -inf.
        linthresh = np.percentile(np.unique(quantity), 1)
        norm = SymLogNorm(linthresh=float(max(linthresh, 1.0e-10 * quantity.max())), base=10)
    else:
        norm = None
    # Sampled data is indexed as quantity(x, y), but matplotlib's imshow
    # expects quantity(y, x).
    if symmetric and not logscale:
        vmax = np.abs(quantity).max()
        image = ax.imshow(
            quantity.T, extent=extent, origin="lower", vmin=-vmax, vmax=vmax, cmap="coolwarm"
        )
    else:
        image = ax.imshow(quantity.T, extent=extent, origin="lower", norm=norm, cmap="gnuplot")
    fig.colorbar(image, aspect=50)
    ax.set_xlim(extent[0], extent[1])
    ax.set_ylim(extent[2], extent[3])
    ax.set_xlabel("R [m]")
    ax.set_ylabel("Z [m]")
    ax.set_title(title)

    return fig


demos_path = Path(__file__).parent
plots_path = demos_path / "plots"
plots_path.mkdir(exist_ok=True)

# sampling range
xl, xu = 4.0, 8.5
zl, zu = -4.5, 4.6
nz = 911
nx = 451
extent = [xl, xu, zl, zu]

# Load and plot equilibrium
try:
    equilibrium = load_equilibrium(iter_jintrac(), "r")
    plot_equilibrium(equilibrium)
    plt.gcf().savefig(plots_path / "equilibrium.png", dpi=200)
    b_field = equilibrium.b_field
except RuntimeError:
    try:
        b_field = load_magnetic_field(iter_jintrac(), "r")
    except RuntimeError:
        b_field = None

# Sample and plot magnetic field
plot_velocity = False
if b_field is not None:
    try:
        xsamp, zsamp, b = samplevector2d(b_field, (xl, xu, nx), (zl, zu, nz))
        b_length = np.sqrt((b * b).sum(2))
        fig = plot_quantity(b[:, :, 0], extent, title="Brad [T]", symmetric=True)
        fig.savefig(plots_path / "brad.png", dpi=200)
        fig = plot_quantity(b[:, :, 1], extent, title="Btor [T]")
        fig.savefig(plots_path / "btor.png", dpi=200)
        fig = plot_quantity(b[:, :, 2], extent, title="Bz [T]", symmetric=True)
        fig.savefig(plots_path / "bz.png", dpi=200)
        radial_vector = np.zeros_like(b)
        radial_vector[:, :, 0] = -b[:, :, 2]
        radial_vector[:, :, 2] = b[:, :, 0]
        radial_vector /= np.sqrt((radial_vector * radial_vector).sum(2))[:, :, None]
        poloidal_vector = np.copy(b)
        poloidal_vector[:, :, 1] = 0
        poloidal_vector /= np.sqrt((poloidal_vector * poloidal_vector).sum(2))[:, :, None]
        plot_velocity = True
    except ValueError:
        print(
            "Unable to sample magnetic field. "
            + "The magnetic field grid does not cover the edge region."
        )

# Load edge plasma
plasma = load_edge_plasma(iter_jintrac(), "r", b_field=b_field)

# Sample and plot electron profiles
xsamp, _, zsamp, ne_plasma = sample3d(
    plasma.electron_distribution.density, (xl, xu, nx), (0, 0, 1), (zl, zu, nz)
)
ne_plasma = ne_plasma.squeeze()
fig = plot_quantity(ne_plasma, extent, title="ne [m-3]", logscale=True)
fig.savefig(plots_path / "edge_ne.png", dpi=200)

te_plasma = sample3d_grid(
    plasma.electron_distribution.effective_temperature, xsamp, [0], zsamp
).squeeze()
fig = plot_quantity(te_plasma, extent, title="Te [eV]", logscale=True)
fig.savefig(plots_path / "edge_te.png", dpi=200)

if plot_velocity:
    electron_velocity = samplevector3d_grid(
        plasma.electron_distribution.bulk_velocity, xsamp, [0], zsamp
    ).squeeze()
    electron_velocity_par = (electron_velocity * b).sum(2) / b_length
    fig = plot_quantity(
        electron_velocity_par,
        extent,
        title="Electron Vpar [m/s]",
        symmetric=True,
    )
    fig.savefig(plots_path / "edge_electron_vpar.png", dpi=200)

# Sample and plot ion and neutral profiles
for species in plasma.composition:
    density = sample3d_grid(species.distribution.density, xsamp, [0], zsamp).squeeze()
    fig = plot_quantity(
        density,
        extent,
        title=f"{species.element.symbol}{species.charge} density [m-3]",
        logscale=True,
    )
    fig.savefig(
        plots_path / f"edge_{species.element.symbol}{species.charge}_density.png",
        dpi=200,
    )

    if species.element.atomic_number == 1:
        temperature = sample3d_grid(
            species.distribution.effective_temperature, xsamp, [0], zsamp
        ).squeeze()
        fig = plot_quantity(
            temperature,
            extent,
            title=f"{species.element.symbol}{species.charge} temperature [eV]",
            logscale=True,
        )
        fig.savefig(
            plots_path / f"edge_{species.element.symbol}{species.charge}_temperature.png",
            dpi=200,
        )

        if plot_velocity:
            velocity = samplevector3d_grid(
                species.distribution.bulk_velocity, xsamp, [0], zsamp
            ).squeeze()

            if species.charge:
                vpar = (velocity * b).sum(2) / b_length
                fig = plot_quantity(
                    vpar,
                    extent,
                    title=f"{species.element.symbol}{species.charge} Vpar [m/s]",
                    symmetric=True,
                )
                fig.savefig(
                    plots_path / f"edge_{species.element.symbol}{species.charge}_vpar.png",
                    dpi=200,
                )
            else:
                vrad = (velocity * radial_vector).sum(2)
                vpol = (velocity * poloidal_vector).sum(2)
                vtor = velocity[:, :, 1]
                fig = plot_quantity(
                    vrad,
                    extent,
                    title=f"{species.element.symbol}{species.charge} Vrad [m/s]",
                    symmetric=True,
                )
                fig.savefig(
                    plots_path / f"edge_{species.element.symbol}{species.charge}_vrad.png",
                    dpi=200,
                )
                fig = plot_quantity(
                    vpol,
                    extent,
                    title=f"{species.element.symbol}{species.charge} Vpol [m/s]",
                    symmetric=True,
                )
                fig.savefig(
                    plots_path / f"edge_{species.element.symbol}{species.charge}_vpol.png",
                    dpi=200,
                )
                fig = plot_quantity(
                    vtor,
                    extent,
                    title=f"{species.element.symbol}{species.charge} Vtor [m/s]",
                    symmetric=True,
                )
                fig.savefig(
                    plots_path / f"edge_{species.element.symbol}{species.charge}_vtor.png",
                    dpi=200,
                )
